import itertools
import numpy as np
from scipy.stats import qmc
from ...runtime import logs
from .. import ParallelSampler


def task(p):
    i, p, name, training = p
    results = lh_sampler.pipeline.run_results(p)
    #If requested, save the data to file
    if lh_sampler.save_name and results.block is not None:
        if training:
            results.block.to_pickle(f"{lh_sampler.save_name}{name}_{i}")
        else:
            results.block.save_to_file(f"{lh_sampler.save_name}{name}_{i}", clobber=True)
    return (results.post, results.prior, results.extra)


class LatinHypercubeSampler(ParallelSampler):
    parallel_output = False
    sampler_outputs = [("prior", float), ("post", float)]
    understands_fast_subspaces = True

    def config(self):
        global lh_sampler
        lh_sampler = self

        self.converged = False
        self.nsample = self.read_ini("nsample", int, 1)
        self.save_name = self.read_ini("save", str, "")
        self.nstep = self.read_ini("nstep", int, -1)

        if self.pipeline.training:
            self.nsample = self.pipeline.nsample
            self.nsample_test = self.pipeline.nsample_test
            self.save_name = self.pipeline.save_name
        else:
            self.nsample_test = 0

        self.sample_points = None
        self.ndone = 0
        self.ndone_train = 0
        self.ndone_test = 0

    def setup_sampling(self):
        #Number of jobs to do at once.
        #Can be taken from the ini file.
        #Otherwise it is set to -1 by default
        if self.nstep==-1:
            #if in parallel mode do a chunk of 4*the number of tasks to do at once
            #chosen arbitrarily.
            if self.pool:
                self.nstep = 4*(self.pool.size-1)
            #if not parallel then just do a single slice through one dimension each chunk
            else:
                self.nstep = self.nsample + self.nsample_test
        
        # If our pipeline allows it we arrange it so that the
        # fast parameters change fastest in the sequence.
        # This is still not optimal for the multiprocessing case
        if self.pipeline.do_fast_slow:
            param_order = self.pipeline.slow_params + self.pipeline.fast_params
        else:
            param_order = self.pipeline.varied_params
        
        # Create a Latin Hypercube sampler
        sampler = qmc.LatinHypercube(d=len(param_order))

        # Generate all the Latin Hypercube samples at once
        lhs = sampler.random(n=self.nsample)

        # Extract the lower and upper bounds for each parameter
        l_bounds = [param.limits[0] for param in param_order]
        u_bounds = [param.limits[1] for param in param_order]

        # Scale the samples to the desired parameter limits
        lhs_scaled = qmc.scale(lhs, l_bounds, u_bounds)
        self.sample_points = iter(lhs_scaled)

        # We create a second LHS if we use the latin hypercube to create samples for
        # ML and / or CosmoPower training
        if self.pipeline.training:
            if self.nsample_test == 0:
                raise ValueError("nsample_test needs to be greater than 0!")
            # Create a Latin Hypercube sampler
            sampler_test = qmc.LatinHypercube(d=len(param_order))

            # Generate all the Latin Hypercube samples at once
            lhs_test = sampler_test.random(n=self.nsample_test)

            # Scale the samples to the desired parameter limits
            # We use the same bounds!
            lhs_scaled_test = qmc.scale(lhs_test, l_bounds, u_bounds)
            self.sample_points_test = iter(lhs_scaled_test)


    def execute(self):
        #First run only:
        if self.sample_points is None:
            self.setup_sampling()

        #Chunk of tasks to do this run through, of size nstep.
        #This advances the self.sample_points forward so it knows
        #that these samples have been done
        samples = list(itertools.islice(self.sample_points, self.nstep))
        if self.pipeline.training:
            samples_test = list(itertools.islice(self.sample_points_test, self.nstep))
        else:
            samples_test = []
        samples_tot = samples + samples_test

        #If there are no samples left then we are done.
        if not samples_tot:
            self.converged=True
            return
        
        #Each job has an index number in case we are saving
        #the output results from each one
        sample_index = np.arange(len(samples)) + self.ndone_train
        sample_name = ["" for _ in sample_index]
        training_flag = [self.pipeline.training for _ in sample_index]
        jobs = list(zip(sample_index, samples, sample_name, training_flag))

        #Actually compute the likelihood results
        if self.pool:
            results = self.pool.map(task, jobs)
        else:
            results = list(map(task, jobs))

        if self.pipeline.training:
            sample_test_index = np.arange(len(samples_test)) + self.ndone_test
            sample_name_test = ["_test" for i in sample_test_index]
            training_flag_test = [self.pipeline.training for i in sample_index]
            jobs_test = list(zip(sample_test_index, samples_test, sample_name_test, training_flag_test))
            if self.pool:
                results_test = self.pool.map(task, jobs_test)
            else:
                results_test = list(map(task, jobs_test))
            self.ndone_test += len(results_test)

        #Update the count
        self.ndone_train += len(results)
        self.ndone = self.ndone_train + self.ndone_test

        #Save the results of the sampling
        for sample, result  in zip(samples, results):
            #Optionally save all the results calculated by each
            #pipeline run to files
            (prob, prior, extra) = result
            #always save the usual text output
            self.output.parameters(sample, extra, prior, prob)
            self.distribution_hints.set_peak(sample, prob)

    def is_converged(self):
        return self.converged



import cosmosis
import pytest
import tempfile

def test_missing_module():
    with pytest.raises(cosmosis.runtime.SetupError) as error_info:
        m = cosmosis.Module("pretend", "./this_file_does_not_exist.py")

    assert "file does not exist" in str(error_info.value)


def test_wrong_path_type():
    with tempfile.NamedTemporaryFile(suffix=".x") as f:
        with pytest.raises(cosmosis.runtime.SetupError) as error_info:
            m = cosmosis.Module("pretend", f.name)

    assert "do not know what kind of module this is" in str(error_info.value)



#coding: utf-8
from .. import ParallelSampler
from ...runtime import logs
import ctypes as ct
import os
import numpy as np
import sys
from ...runtime.utils import mkdir
from ...output import TextColumnOutput
import warnings

prior_type = ct.CFUNCTYPE(None, 
    ct.POINTER(ct.c_double),  #hypercube
    ct.POINTER(ct.c_double),  #physical
    ct.c_int,   #ndim
)

loglike_type = ct.CFUNCTYPE(ct.c_double, 
    ct.POINTER(ct.c_double),  #physical
    ct.c_int,   #ndim
    ct.POINTER(ct.c_double),  #derived
    ct.c_int,   #nderived
)


dumper_type = ct.CFUNCTYPE(None, #void
    ct.c_int,  #ndead
    ct.c_int,  #nlive
    ct.c_int,  #npars
    ct.POINTER(ct.c_double),   #live
    ct.POINTER(ct.c_double),   #dead
    ct.POINTER(ct.c_double),   #logweights
    ct.c_double,   #logZ
    ct.c_double,   #logZerr
)


polychord_args = [
    loglike_type,           #loglike,
    prior_type,             #prior,
    dumper_type,            #dumper,
    ct.c_int,               #nlive
    ct.c_int,               #nrepeats
    ct.c_int,               #nprior
    ct.c_bool,              #do_clustering
    ct.c_int,               #feedback
    ct.c_double,            #precision_criterion
    ct.c_double,            #logzero
    ct.c_int,               #max_ndead
    ct.c_double,            #boost_posterior
    ct.c_bool,              #posteriors
    ct.c_bool,              #equals
    ct.c_bool,              #cluster_posteriors
    ct.c_bool,              #write_resume 
    ct.c_bool,              #write_paramnames
    ct.c_bool,              #read_resume
    ct.c_bool,              #write_stats
    ct.c_bool,              #write_live
    ct.c_bool,              #write_dead
    ct.c_bool,              #write_prior
    ct.c_double,            #compression_factor
    ct.c_int,               #nDims
    ct.c_int,               #nDerived 
    ct.c_char_p,            #base_dir
    ct.c_char_p,            #file_root
    ct.c_int,               #nGrade
    ct.POINTER(ct.c_double),#grade_frac
    ct.POINTER(ct.c_int),   #grade_dims
    ct.c_int,               #n_nlives
    ct.POINTER(ct.c_double),#loglikes
    ct.POINTER(ct.c_int),   #nlives
    ct.c_int,               #seed
]


POLYCHORD_SECTION='polychord'


class PolychordSampler(ParallelSampler):
    parallel_output = False
    sampler_outputs = [("prior", float), ("like", float), ("post", float), ("weight", float)]
    supports_smp=False
    internal_resume = True
    understands_fast_subspaces = True

    def config(self):
        if self.pool:
            libname = "libchord_mpi.so"
        else:
            libname = "libchord.so"

        dirname = os.path.split(__file__)[0]
        libname = os.path.join(dirname, "polychord_src", libname)
            
        try:
            libchord = ct.cdll.LoadLibrary(libname)
        except Exception as error:
            sys.stderr.write("PolyChord could not be loaded.\n")
            sys.stderr.write("This may mean an MPI compiler was not found to compile it,\n")
            sys.stderr.write("or that some other error occurred.  More info below.\n")
            sys.stderr.write(str(error)+'\n')
            sys.exit(1)

        self._run = libchord.polychord_c_interface
        self._run.restype=None
        self._run.argtypes = polychord_args
        self.converged=False

        self.ndim = len(self.pipeline.varied_params)
        # We save the prior as well as the other derived params
        self.nderived = self.pipeline.number_extra + 1

        #Required options
        self.live_points    = self.read_ini("live_points", int, 100)

        #Output and feedback options
        self.feedback               = self.read_ini("feedback", int, 1)
        self.resume                 = self.read_ini("resume", bool, False)
        self.polychord_outfile_root = self.read_ini("polychord_outfile_root", str, "")
        self.base_dir               = self.read_ini("base_dir", str, ".")
        self.compression_factor     = self.read_ini("compression_factor", float, np.exp(-1))

        if len(self.polychord_outfile_root)>299:
            raise ValueError("Polychord parameter 'polychord_outfile_root'"
                " cannot be longer than 299s characters.")

        if '/' in self.polychord_outfile_root:
            warnings.warn("\n\nWARNING: / found in polychord_outfile_root\n"
                "In polychord it is better to set base_dir to the directory and \n"
                "have polychord_outfile_root not include the directory.\n"
                "Otherwise you may have to mkdir some directories yourself.\n"
                "This may cause problems below.\n"
                )

        #General run options
        self.max_iterations = self.read_ini("max_iterations", int, -1)
        self.num_repeats = self.read_ini("num_repeats", int, 0)
        self.nprior = self.read_ini("nprior", int, -1)
        self.random_seed = self.read_ini("random_seed", int, -1)
        self.tolerance   = self.read_ini("tolerance", float, 0.1)
        self.log_zero    = self.read_ini("log_zero", float, -1e6)
        self.boost_posteriors = self.read_ini("boost_posteriors", float, 0.0)
        self.weighted_posteriors = self.read_ini("weighted_posteriors", bool, True)
        self.equally_weighted_posteriors = self.read_ini("equally_weighted_posteriors", bool, True)       
        self.cluster_posteriors = self.read_ini("cluster_posteriors", bool, True)

        self.fast_fraction    = self.read_ini("fast_fraction", float, 0.5)

        if self.output:
            def dumper(ndead, nlive, npars, live, dead, logweights, log_z, log_z_err):
                logs.overview("Saving %d samples" % ndead)
                self.output_params(ndead, nlive, npars, live, dead, logweights, log_z, log_z_err)
            self.wrapped_output_logger = dumper_type(dumper)
        else:
            def dumper(ndead, nlive, npars, live, dead, logweights, log_z, log_z_err):
                pass
            self.wrapped_output_logger = dumper_type(dumper)

        def prior(cube, theta, ndim):
            cube_vector = np.array([cube[i] for i in range(ndim)])
            if self.pipeline.do_fast_slow:
                cube_vector = self.reorder_slow_fast(cube_vector)
            try:
                theta_vector = self.pipeline.denormalize_vector_from_prior(cube_vector) 
            except ValueError:
                # Polychord sometimes seems to propose outside the prior.
                # Just give terrible parameters when that happens.
                theta_vector = np.repeat(-np.inf, ndim)
            for i in range(ndim):
                theta[i] = theta_vector[i]

        self.wrapped_prior = prior_type(prior)

        def likelihood(theta, ndim, phi, nderived):
            theta_vector = np.array([theta[i] for i in range(ndim)])

            if np.any(~np.isfinite(theta_vector)):
                return -np.inf

            try:
                r = self.pipeline.run_results(theta_vector)

                # Extract the derived parameters
                for i in range(nderived-1):
                    phi[i] = r.extra[i]
                # And we also handle the prior here
                phi[nderived-1] = r.prior

            except KeyboardInterrupt:
                raise sys.exit(1)

            return r.like
        self.wrapped_likelihood = loglike_type(likelihood)

    def reorder_slow_fast(self, x):
        y = np.zeros_like(x)
        ns = self.pipeline.n_slow_params
        y[self.pipeline.slow_param_indices] = x[0:ns]
        y[self.pipeline.fast_param_indices] = x[ns:]
        return y



    def worker(self):
        self.sample()

    def execute(self):
        self.log_z = 0.0
        self.log_z_err = 0.0


        self.sample()

        self.output.final("log_z", self.log_z)
        self.output.final("log_z_error", self.log_z_err)

    def sample(self):

        if self.pipeline.do_fast_slow and (self.pipeline.n_fast_params > 0):
            logs.overview("Using two grades of parameter speed in polychord.")
            n_grade = 2
        elif self.pipeline.do_fast_slow:
            logs.warning("You asked for fast/slow, but there were no fast parameters, so "
                  "I have switched off Polychord's fast/slow mechanism to avoid a hang")
            n_grade = 1
        else:
            logs.overview("Using a single grade of parameter speeds in polychord.")
            n_grade = 1

        grade_dims = (ct.c_int*n_grade)()
        grade_frac = (ct.c_double*n_grade)()


        if n_grade > 1:
            if self.pipeline.n_slow_params == 0:
                raise ValueError("All your parameters have been classified as fast.  This will now work in polychord.  Change your fast/slow settings.")

            grade_dims[0] = self.pipeline.n_slow_params
            grade_dims[1] = self.pipeline.n_fast_params
            grade_frac[0] = 1 - self.fast_fraction
            grade_frac[1] = self.fast_fraction
            logs.overview("Telling Polychord to spend fraction {} if its time in the fast subspace (adjust with fast_fraction option)".format(self.fast_fraction))
        else:
            grade_dims[0] = self.pipeline.nvaried
            grade_frac[0] = 1.0


        n_nlives = 0
        loglikes = (ct.c_double*n_nlives)()
        nlives = (ct.c_int*n_nlives)()

        base_dir = self.base_dir.encode('ascii')
        polychord_outfile_root = self.polychord_outfile_root.encode('ascii')
        output_to_file = len(polychord_outfile_root) > 0

        if output_to_file:
            mkdir(self.base_dir)
            mkdir(os.path.join(self.base_dir, "clusters"))
            
        if self.num_repeats == 0:
            num_repeats = 3 * grade_dims[0]
            logs.overview("Polychord num_repeats = {}  (3 * n_slow_params [{}])".format(num_repeats, grade_dims[0]))
        else:
            num_repeats = self.num_repeats
            logs.overview("Polychord num_repeats = {}  (from parameter file)".format(num_repeats))

        self._run(
                self.wrapped_likelihood,      #loglike,
                self.wrapped_prior,           #prior,
                self.wrapped_output_logger,   #dumper,
                self.live_points,             #nlive
                num_repeats,                  #nrepeats
                self.nprior,                  #nprior
                True,                         #do_clustering
                self.feedback,                #feedback
                self.tolerance,               #precision_criterion
                self.log_zero,                #logzero
                self.max_iterations,          #max_ndead
                self.boost_posteriors,         #boost_posterior
                self.weighted_posteriors,     #posteriors
                self.equally_weighted_posteriors, #equals
                self.cluster_posteriors,      #cluster_posteriors
                True,                         #write_resume  - always
                False,                        #write_paramnames
                self.resume,                  #read_resume
                output_to_file,  #write_stats
                output_to_file,  #write_live
                output_to_file,  #write_dead
                output_to_file,  #write_prior
                self.compression_factor,      #compression_factor
                self.ndim,                    #nDims
                self.nderived,                #nDerived 
                base_dir,           #base_dir
                polychord_outfile_root,  #file_root
                n_grade,                      #nGrade
                grade_frac,                   #grade_frac
                grade_dims,                   #grade_dims
                n_nlives,                     #n_nlives
                loglikes,                     #loglikes
                nlives,                       #nlives
                self.random_seed,             #seed
                )

        self.converged = True

        if self.is_master() and self.boost_posteriors:
            self.make_boosted_posterior_file()

    def make_boosted_posterior_file(self):
        # two cases where there is nothing to boost
        if not self.polychord_outfile_root:
            return
        if not self.boost_posteriors:
            return

        old = self.output

        boosted_filename = os.path.join(self.base_dir, self.polychord_outfile_root + ".txt")

        # This is a horribly awkward hack
        if isinstance(self.output, TextColumnOutput):
            # make sure everything is written out already
            old.flush()
            
            new_output_filename = old.filename_base + "_boosted.txt"
            main_output_info = TextColumnOutput.load_from_options({"filename":old._filename, "delimiter":old.delimiter})

            _, _, metadata, comments, final_metadata = main_output_info
            metadata = metadata[0]
            comments = comments[0]
            final_metadata = final_metadata[0]

            # make a new output object
            new = TextColumnOutput.from_options({"filename":new_output_filename})

            # write the main bits of metadata from the output file
            new.metadata("sampler", "polychord")
            new.comment("NOTE: Boosted posterior file")
            self.write_header(new)
            for md in metadata:
                new.metadata(md, metadata[md])
            for c in comments:
                new.comment(c)
            
            nsample = 0
            # write the parameter rows from the old file
            with open(boosted_filename, "r")  as f:
                for line in f:
                    values = [float(x) for x in line.split()]
                    weight = values[0]
                    like = -0.5 * values[1]
                    params = values[2:]
                    prior = params[-1]
                    post = like + prior
                    new.parameters(params, like, post, weight)
                    nsample += 1

            metadata = old._final_metadata.copy()
            metadata["nsample"] = nsample
            # This hasn't been written yet because the original output file has not been closed.
            # so we have to use the saved one in the old object. The * is because a comment is also included
            for key, value in metadata.items():
                if isinstance(value, tuple):
                    new.final(key, *value)
                else:
                    new.final(key, value)
            new.final("log_z", self.log_z)
            new.final("log_z_error", self.log_z_err)

            # fake these first two, and add the completion marker.
            new.final("evaluations", 0)
            new.final("successes", 0)
            new.final("complete", 1)
            new.close()



    def output_params(self, ndead, nlive, npars, live, dead, logweights, log_z, log_z_err):
        # Polychord repeats output, but with changed weights, so reset to the start
        # of the chain to overwrite them.
        self.output.reset_to_chain_start()
        self.log_z = log_z
        self.log_z_err = log_z_err
        data = np.array([dead[i] for i in range(npars*ndead)]).reshape((ndead, npars))
        logw = np.array([logweights[i] for i in range(ndead)])
        for row, w in zip(data,logw):
            params = row[:self.ndim]
            extra_vals = row[self.ndim:self.ndim+self.nderived-1]
            prior = row[self.ndim+self.nderived-1]
            birth_like = row[self.ndim+self.nderived]
            like = row[self.ndim+self.nderived+1]
            importance = np.exp(w)
            post = like + prior
            self.output.parameters(params, extra_vals, prior, like, post, importance)

        # priors + likes
        posts = data[:, self.ndim+self.nderived-1] + data[:, self.ndim+self.nderived+1]
        self.distribution_hints.set_from_sample(data[:, :self.ndim], posts, log_weights=logw)

        self.output.final("nsample", ndead)
        self.output.flush()

    def is_converged(self):
        return self.converged

# legacy alias
PolyChordSampler = PolychordSampler



import os
import sys
import numpy as np

class Loadable(object):
    """ A series of methods to allow creating instances from python files"""
    @classmethod
    def subclasses_in_file(cls, filepath):
        dirname, filename = os.path.split(filepath)
        # allows .pyc and .py modules to be used
        impname, ext = os.path.splitext(filename)
        sys.path.insert(0, dirname)
        try:
            library = __import__(impname)
        except Exception as error:
            print("Could not find/load extension file %s:" % filepath)
            print("ERROR:", error)
            return []

        subclasses = []
        for x in dir(library):
            x = getattr(library, x)
            if type(x) == type and issubclass(x, cls) and not x is cls:
                subclasses.append(x)
        return subclasses


    @classmethod
    def instances_from_file(cls, filename, *args, **kwargs):
        subclasses = cls.subclasses_in_file(filename)
        return [s(*args, **kwargs) for s in subclasses]




class PostProcessorElement(Loadable):
    def __init__(self, data_source, **options):
        super(PostProcessorElement,self).__init__()
        self.source = data_source
        self.options = {}
        self.options.update(options)

    def reduced_col(self, name, stacked=True):
        return self.source.reduced_col(name, stacked=stacked)

    def get_output(self, name):
        return self.source.outputs.get(name)

    def filename(self, ftype, base, *bases):
        if bases:
            base = base + "_" + ("_".join(bases))
        output_dir = self.options.get("outdir", ".")
        prefix=self.options.get("prefix","")
        if prefix: prefix+="_"
        return "{0}/{1}{2}.{3}".format(output_dir, prefix, base, ftype)

    def set_output(self, name, value):
        if name in self.source.outputs:
            raise ValueError("Two different postprocessors tried to create the same named file - use get_output to re-open an existing file")
        self.source.outputs[name] = value

    def run(self):
        print("I do not know how to produce some results for this kind of data")
        return []

    def finalize(self):
        pass

class MCMCPostProcessorElement(PostProcessorElement):
    def posterior_sample(self):
        return self.source.posterior_sample()

class WeightedMCMCPostProcessorElement(MCMCPostProcessorElement):
    def weight_col(self):
        return self.source.weight_col()

class MultinestPostProcessorElement(MCMCPostProcessorElement):
    def weight_col(self):
        return self.source.weight_col()

class PolychordPostProcessorElement(MultinestPostProcessorElement):
    pass



import os
import sys
import argparse
import subprocess

parser = argparse.ArgumentParser(description="print commands that set up the cosmosis env")
cosmosis_src_dir = os.path.split(__file__)[0]
parser.add_argument("--source", default=cosmosis_src_dir)
parser.add_argument("--no-omp", action='store_false', dest='omp', help='Switch off OpenMP')
parser.add_argument("--omp", action='store_true', dest='omp', help='Switch on OpenMP (default)')
parser.add_argument("--debug", action='store_true', dest='debug', help='Switch on debug mode')
parser.add_argument("--no-debug", action='store_false', dest='debug', help='Switch on debug mode')
parser.add_argument("--no-conda", action='store_false', dest='conda', help='Switch off conda flags, even if conda env is found')
parser.add_argument("--brew", action='store_true', help='Print commands for homebrew with clang')
parser.add_argument("--brew-gcc", action='store_true', help='Print commands for homebrew with gcc')
parser.add_argument("--ports", action='store_true', help='Print commands for macports')
parser.add_argument("--automate-conda-setup", action='store_true', help='Automatically set up cosmosis when activating the environment from now on')

def homebrew_gfortran_libs():
    s = subprocess.run('gfortran -print-search-dirs', shell=True, capture_output=True)
    if s.returncode:
        return ""
    lines = s.stdout.decode().split("\n")

    for line in lines:
        if line.startswith("libraries:"):
            break
    else:
        return ""
    try:
        libdir = line.split("=")[1].split(":")[-1]
    except:
        return ""

    return f"-L {libdir}"

def homebrew_gcc_commands():
    s = subprocess.run('brew list --versions gcc', shell=True, capture_output=True)
    version = s.stdout.decode().split()[1].split('.')[0]
    return [
        f"export CC=gcc-{version}",
        f"export CXX=g++-{version}",
        f"export FC=gfortran-{version}",
        "export MPIFC=mpif90",
        "export COSMOSIS_ALT_COMPILERS=1",
    ]



def generate_commands(cosmosis_src_dir, debug=False, omp=True, brew=False, brew_gcc=False, conda=True, ports=False):
    conda = conda and ("CONDA_PREFIX" in os.environ)

    commands = [
        f"export COSMOSIS_SRC_DIR=\"{cosmosis_src_dir}\"",
        "export COSMOSIS_ALT_COMPILERS=1",
    ]

    if not brew:
        commands += [
            "export LIBRARY_PATH=$LIBRARY_PATH:$COSMOSIS_SRC_DIR/datablock",
            "export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$COSMOSIS_SRC_DIR/datablock",
    ]

    if brew:
        commands += [
            "export GSL_LIB=/usr/local/lib",
            "export GSL_INC=/usr/local/include",
            "export FFTW_LIBRARY=/usr/local/lib",
            "export FFTW_INCLUDE_DIR=/usr/local/include",
            "export LAPACK_LINK='-L /usr/local/opt/openblas/lib/ -l lapack'",
            "export LAPACK_LIB=/usr/local/opt/openblas/lib/",
            "export CFITSIO_LIB=/usr/local/lib",
            "export CFITSIO_INC=/usr/local/include",
        ]

        if brew_gcc:
            commands += homebrew_gcc_commands()
        else:
            commands += [
                f"export CC=clang",
                f"export CXX=clang++",
                f"export FC=gfortran",
                "export MPIFC=mpif90",
                "export COSMOSIS_ALT_COMPILERS=1",
            ]

    elif ports:
        commands += [
            'export GSL_INC=/opt/local/include',
            'export GSL_LIB=/opt/local/lib',
            'export CFITSIO_LIB=/opt/local/lib',
            'export CFITSIO_INC=/opt/local/include',
            'export FFTW_LIBRARY=/opt/local/lib',
            'export FFTW_INCLUDE_DIR=/opt/local/include',
            'export LAPACK_LINK="-L/opt/local/lib -llapack -lblas"',
            'export LAPACK_LIB=/opt/local/lib',
            'export CXX=/opt/local/bin/g++',
            'export CC=/opt/local/bin/gcc',
            'export FC=/opt/local/bin/gfortran',
            'export MPICC=mpicc',
            'export MPICXX=mpicxx',
            'export MPIFC=mpifort',
        ]

    elif conda:
        commands += [
            'export GSL_LIB=$CONDA_PREFIX/lib',
            'export GSL_INC=$CONDA_PREFIX/include',
            'export FFTW_LIBRARY=$CONDA_PREFIX/lib',
            'export FFTW_INCLUDE_DIR=$CONDA_PREFIX/include',
            'export LAPACK_LINK="-L$CONDA_PREFIX/lib -llapack"',
            'export LAPACK_LIB="$CONDA_PREFIX/lib"',
            'export CFITSIO_LIB=$CONDA_PREFIX/lib',
            'export CFITSIO_INC=$CONDA_PREFIX/include',
            ]

    if omp:
        commands.append("export COSMOSIS_OMP=1")
        
    if debug:
        commands.append("COSMOSIS_DEBUG=1")

    return commands

def automate_conda_setup(cosmosis_src_dir, cmds, conda):
    if not (conda and "CONDA_PREFIX" in os.environ):
        print("Error: --automate-conda-setup requires a conda environment to be active", file=sys.stderr)
        sys.exit(1)

    activate_dir = os.path.join(os.environ["CONDA_PREFIX"], "etc", "conda", "activate.d")
    os.makedirs(activate_dir, exist_ok=True)
    activate_script_path =  os.path.join(activate_dir, "activate_cosmosis.sh")

    with open(activate_script_path, 'w') as f:
        for cmd in cmds:
            f.write(cmd + "\n")

    deactivate_dir = os.path.join(os.environ["CONDA_PREFIX"], "etc", "conda", "deactivate.d")
    os.makedirs(deactivate_dir, exist_ok=True)
    deactivate_script_path = os.path.join(deactivate_dir, "deactivate_cosmosis.sh")

    with open(deactivate_script_path, 'w') as f:
        f.write("unset COSMOSIS_SRC_DIR\n")
        f.write("unset COSMOSIS_ALT_COMPILERS\n")
        f.write("unset GSL_LIB\n")
        f.write("unset GSL_INC\n")
        f.write("unset FFTW_LIBRARY\n")
        f.write("unset FFTW_INCLUDE_DIR\n")
        f.write("unset LAPACK_LINK\n")
        f.write("unset LAPACK_LIB\n")
        f.write("unset CFITSIO_LIB\n")
        f.write("unset CFITSIO_INC\n")
        f.write("unset COSMOSIS_OMP\n")
        f.write("unset COSMOSIS_DEBUG\n")

        v = f"{cosmosis_src_dir}/datablock"
        # remove datablock from LIBRARY_PATH and LD_LIBRARY_PATH
        f.write('export LIBRARY_PATH=$(echo $LIBRARY_PATH | tr ":" "\\n" | grep -v "{}" | tr "\\n" ":")\n'.format(v))
        f.write('export LD_LIBRARY_PATH=$(echo $LD_LIBRARY_PATH | tr ":" "\\n" | grep -v "{}" | tr "\\n" ":")\n'.format(v))

    print(f"CosmoSIS will automatically configure when you activate the environment from now on", file=sys.stderr)


if __name__ == '__main__':
    args = parser.parse_args()
    cmds = generate_commands(args.source, debug=args.debug, omp=args.omp, conda=args.conda, brew=args.brew or args.brew_gcc, brew_gcc=args.brew_gcc, ports=args.ports)
    if args.automate_conda_setup:
        automate_conda_setup(args.source, cmds, args.conda)
    print("; ".join(cmds))




from .block import DataBlock, BlockError
from .import section_names as names
import os

try:
	from .lib import enable_cosmosis_segfault_handler
except:
	_allow_unbuilt_import = os.environ.get("COSMOSIS_ALLOW_UNBUILT_IMPORT", "0") == "1"
	if not _allow_unbuilt_import:
		raise



from .output_base import OutputBase

class NullOutput(OutputBase):
    _aliases = ["none"]
    def _write_parameters(self, params):
        pass

    def _write_metadata(self, key, value, comment):
        pass

    def _write_comment(self, comment):
        pass

    def _write_final(self, key, value, comment):
        pass

    @classmethod
    def from_options(cls, options, resume=False):
        if resume:
            raise ValueError("Cannot resume from null output")
        return cls()

    @classmethod
    def load_from_options(cls, options):
        raise ValueError("No output was saved from this run")



from .text_output import TextColumnOutput
from .cosmomc_output import CosmoMCOutput
from .null_output import NullOutput
from .fits_output import FitsOutput
from .in_memory_output import InMemoryOutput
from .astropy_output import AstropyOutput
from .output_base import output_registry, OutputBase


def output_from_options(options, resume=False):
	# figure out the type of output required
	format = options.get('format', 'text')
	if format not in output_registry:
		known = '\n'.join('    %s - %s' % (f,output_registry[f].__name__) for f in output_registry)
		message = """
I do not know what format you meant by '%s' in the [output] ini file section.
I know about these format names:
%s
		""" % (format, known)
		print(message)
		raise KeyError("Unknown format")

	output_class = output_registry[format]

	return output_class.from_options(options,resume)

def input_from_options(options):
    format = options['format']
    if format not in output_registry:
        known = '\n'.join('    %s - %s' % (f,output_registry[f].__name__) for f in output_registry)
        message = """
I do not know what format you meant by '%s' in the [output] ini file section.
I know about these format names:
%s
        """ % (format, known)
        print(message)
        raise KeyError("Unknown format")

    output_class = output_registry[format]
    return output_class.load_from_options(options)

def test():
	from . import test_output
	test_output.test_text()



from .runtime import Inifile, MPIPool
from .main import run_cosmosis, run_under_debugger
from .runtime.utils import underline
import time
import os
import yaml
import sys
import warnings
import subprocess
import contextlib

class UniqueKeyLoader(yaml.SafeLoader):
    """
    This is a YAML loader that raises an error if there are duplicate keys.

    It is based on the discussion here:
    https://gist.github.com/pypt/94d747fe5180851196eb
    """
    def construct_mapping(self, node, deep=False):
        mapping = set()
        for key_node, _ in node.value:
            if ":merge" in key_node.tag:
                continue
            key = self.construct_object(key_node, deep=deep)
            if key in mapping:
                raise ValueError(f"Duplicate {key} key found in YAML.")
            mapping.add(key)
        return super().construct_mapping(node, deep)

def load_yaml(stream):
    """
    Load YAML markup from a stream, with a check for duplicate keys.

    Parameters
    ----------
    stream : file
        The open file-like object to load from

    Returns
    -------
    data : object
        The dict or list loaded from the YAML file
    """
    return yaml.load(stream, Loader=UniqueKeyLoader)


def pipeline_after(params, after_module, modules):
    """
    Insert item(s) in the pipeline after the given module.

    Parameters
    ----------
    params : Inifile
        The parameter file to modify
    after_module : str
        The name of the module after which to insert the new modules
    modules : list
        The list of modules to insert

    Returns
    -------
    None
    """
    pipeline = params['pipeline', 'modules']
    pipeline = pipeline.split()
    if isinstance(modules, str):
        modules = [modules]

    i = pipeline.index(after_module)

    for m in modules[::-1]:
        pipeline.insert(i+1, m)
    params.set('pipeline', 'modules', ' '.join(pipeline))

def pipeline_before(params, before_module, modules):
    """
    Insert item(s) in the pipeline before the given module.

    If before_module is "$" then the item is inserted at the end

    Parameters
    ----------
    params : Inifile
        The parameter file to modify
    before_module : str 
        The name of the module before which to insert the new modules
    modules : list
        The list of modules to insert
    """
    if isinstance(modules, str):
        modules = [modules]

    pipeline = params['pipeline', 'modules']
    pipeline = pipeline.split()
    i = pipeline.index(before_module)

    for m in modules[::-1]:
        pipeline.insert(i, m)
    params.set('pipeline', 'modules', ' '.join(pipeline))

def pipeline_replace(params, original_module, modules):
    """
    Replace an item in the pipeline with one or more new items.

    If the original module is not found then this will fail.

    Parameters
    ----------
    params : Inifile
        The parameter file to modify
    original_module : str
        The name of the module to replace
    modules : list
        The list of modules to insert

    Returns
    -------
    None
    """
    pipeline_after(params, original_module, modules)
    pipeline_delete(params, [original_module])

def pipeline_delete(params, modules):
    """
    Delete one or more items in the pipeline.

    Parameters
    ----------
    params : Inifile
        The parameter file to modify
    modules : list
        The list of modules to delete

    Returns
    -------
    None
    """
    if isinstance(modules, str):
        modules = [modules]
    pipeline = params['pipeline', 'modules']
    pipeline = pipeline.split()

    for m in modules:
        pipeline.remove(m)
    params.set('pipeline', 'modules', ' '.join(pipeline))

def pipeline_append(params, modules):
    """
    Append item(s) to the end of the pipeline.

    Parameters
    ----------
    params : Inifile
        The parameter file to modify
    modules : list
        The list of modules to append
    
    Returns
    -------
    None
    """
    pipeline = params['pipeline', 'modules']
    pipeline = pipeline.split()

    for m in modules:
        pipeline.append(m)
    params.set('pipeline', 'modules', ' '.join(pipeline))

def pipeline_prepend(params, modules):
    """
    Prepend item(s) to the start of the pipeline.

    Parameters
    ----------
    params : Inifile
        The parameter file to modify
    modules : list

    Returns
    -------
    """
    pipeline = params['pipeline', 'modules']
    pipeline = pipeline.split()

    for m in modules[::-1]:
        pipeline.insert(0, m)
    params.set('pipeline', 'modules', ' '.join(pipeline))


def apply_update(ini, update, is_params=False):
    """
    Apply an update to a parameters, value, or priors configuration.

    Parameters
    ----------
    ini : Inifile
        The values or priors file to update
    update : str
        The update to apply. This should be of the form:
        section.option=value
        sampler=sampler_name
        del section.option
        del section
    """
    if "=" in update:
        keys, value = update.split("=", 1)
        keys = keys.strip()
        value = value.strip()
        if keys == "sampler":
            if not is_params:
                raise ValueError("You can only set the sampler in the parameters file")
            ini.set("runtime", "sampler", value.strip())
        else:
            section, option = keys.split(".", 1)
            section = section.strip()
            option = option.strip()
            if not ini.has_section(section) and section != "DEFAULT":
                ini.add_section(section)
            ini.set(section, option, value)
    elif update.startswith("del"):
        cmd, keys = update.split(maxsplit=1)
        if cmd not in ["del", "delete"]:
            raise ValueError(f"Unknown command {cmd}")
        if "." in keys:
            section, option = keys.split('.', 1)
            section = section.strip()
            option = option.strip()
            ini.remove_option(section, option)
        else:
            keys = keys.strip().split()
            if len(keys) > 1:
                raise ValueError("Can only delete one section at a time")
            else:
                # delete entire section
                ini.remove_section(keys[0])
    else:
        raise ValueError(f"Unknown update {update}")

def apply_updates(ini, updates, is_params=False):
    """
    Apply a list of updates to a parameters, values, or priors configuration.

    Parameters
    ----------
    ini : Inifile
        The values or priors file to update
    updates : list
        See apply_update for details.
    
    Returns
    -------
    None
    """
    for update in updates:
        try:
            apply_update(ini, update, is_params=is_params)
        except:
            raise ValueError(f"Malformed update: {update}")
        

def apply_pipeline_update(ini, update):
    """
    Apply an update to the list of modules forming the pipeline.
    
    Available actions are:
    - after <module> <new_module> [<new_module> ...]
    - before <module> <new_module> [<new_module> ...]
    - replace <module> <new_module> [<new_module> ...]
    - del <module> [<module> ...]
    - append <new_module> [<new_module> ...]
    - prepend <new_module> [<new_module> ...]

    Parameters
    ----------
    ini : Inifile
        The parameters file to update
    update : list
        The update to apply.  This should be a list of strings
        of the form [action, module, *values]
        values can be a single string or a list of strings,
        which will be joined with spaces.

    """
    action = update[0]
    if action == "after":
        pipeline_after(ini, update[1], update[2:])
    elif action == "before":
        pipeline_before(ini, update[1], update[2:])
    elif action == "replace":
        pipeline_replace(ini, update[1], update[2:])
    elif action == "del" or action == "delete":
        pipeline_delete(ini, update[1:])
    elif action == "append":
        pipeline_append(ini, update[1:])
    else:
        raise ValueError(f"Unknown pipeline action {action}")


def apply_pipeline_updates(params, pipeline_updates):
    """
    Apply a list of pipeline updates.

    Parameters
    ----------
    params : Inifile
        The parameters file to update
    pipeline_updates : list
        The list of updates to apply.  Each update should be a list of strings.
        See apply_pipeline_update for details.

    Returns
    -------
    None
    """
    for update in pipeline_updates:
        apply_pipeline_update(params, update.split())



@contextlib.contextmanager
def temporary_environment(env):
    """
    Temporary set the environment variables in the given dictionary.
    Once the with block is exited the environment is restored to its
    original state.

    Parameters
    ----------
    env : dict
        A dictionary of environment variables to set

    Returns
    -------
    None
    """
    original_environment = os.environ.copy()
    try:
        os.environ.update(env)
        yield
    finally:
        os.environ.clear()
        os.environ.update(original_environment)

def set_output_dir(params, name, output_dir, output_name):
    """
    Modify a parameters file to set the output directory and file names to be
    based on the run name and output directory.

    Parameters
    ----------
    params : Inifile
        The parameters file to update
    name : str
        The name of the run
    output_dir : str
        The output directory to use
    output_name : str
        The format string to use to generate the output file name, using {name}
        for the run name.
        
    Returns
    -------
    None
    """
    # Always override the output file and various other auxiliary files
    if not params.has_section("output"):
        params.add_section("output")
    if not params.has_section("test"):
        params.add_section("test")

    output_name_base = output_name.format(name=name)
    output_name = output_name_base + ".txt"
    params.set("output", "filename", os.path.join(output_dir, output_name))
    params.set("test", "save_dir", os.path.join(output_dir, output_name_base))

    # Polychord and Multinest have their own output files. They are handled slightly
    # differently. It would be nicer to put all these files in specific directories
    # instead of just using specific base names, and I will do that at the next breaking
    # change, but right now doing that would mean I had to change the multinest sampler
    # to create the directories, which would be backwards incompatible.
    if params.has_section("multinest"):
        params.set("multinest", "multinest_outfile_root", os.path.join(output_dir, f"{output_name_base}.multinest"))

    if params.has_section("polychord"):
        params.set("polychord", "polychord_outfile_root", f"{output_name_base}.polychord")
        params.set("polychord", "base_dir", output_dir)

def build_run(name, run_info, runs, components, output_dir, submission_info, output_name="{name}"):
    """
    Generate a dictionary specifying a CosmoSIS run from a run_info dictioary.

    Parameters
    ----------
    name : str
        The name of the run
    run_info : dict
        A dictionary specifying the run.  This can have the following keys:
        - base : the name of the base parameter file OR
        - parent : the name of the parent run to use as a base (one of these is required)
        - params : a list of updates to apply to the parameters file (optional)
        - values : a list of updates to apply to the values file (optional)
        - priors : a list of updates to apply to the priors file (optional)
        - pipeline : a list of updates to apply to the pipeline (optional)
        - components : a list of components to include (optional)
        - env : a dictionary of environment variables to set (optional)
    runs : dict
        A dictionary of previously built runs
    components : dict
        A dictionary of previously built components
    output_dir : str
        The output directory to use
    submission_info : dict
        A dictionary of submission information
    output_name : str
        The format string to use to generate the output file name, using {name}
        for the run name.

    Returns
    -------
    """
    run = run_info.copy()

    # We want to delay expanding environment variables so that child runs
    # have a chance to override them. So we set no_expand_vars=True on all of these
    if "base" in run_info:
        params = Inifile(run_info["base"], print_include_messages=False, no_expand_vars=True)
    elif "parent" in run_info:
        try:
            parent = runs[run_info["parent"]]
        except KeyError:
            warnings.warn(f"Run {name} specifies parent {run_info['parent']} but there is no run with that name yet")
            return None
        params = Inifile(parent["params"], print_include_messages=False, no_expand_vars=True)
    else:
        warnings.warn(f"Run {name} specifies neither 'parent' nor 'base' so is invalid")
        return None
    
    # Build environment variables
    # These are inherited from the parent run, if there is one,
    # and then updated with any specific to this run, which can overwrite.
    # env vars are only applied right at the end when all runs are collected
    if "parent" in run_info:
        env_vars = parent["env"].copy()
    else:
        env_vars = {}
    env_vars.update(run_info.get("env", {}))
    run["env"] = env_vars

    # Build values file, which is mandatory
    if "parent" in run_info:
        values = Inifile(parent["values"], print_include_messages=False, no_expand_vars=True)
    else:
        values_file = params.get('pipeline', 'values')
        values = Inifile(values_file, print_include_messages=False, no_expand_vars=True)

    # Build optional priors file
    if "parent" in run_info:
        priors = Inifile(parent["priors"], print_include_messages=False, no_expand_vars=True)
    elif "priors" in params.options("pipeline"):
        priors_file = params.get('pipeline', 'priors')
        priors = Inifile(priors_file, print_include_messages=False, no_expand_vars=True)
    else:
        priors = Inifile(None, no_expand_vars=True)

    # Make a list of all the modifications to be applied
    # to the different bits of this pipeline

    param_updates = []
    value_updates = []
    prior_updates = []
    pipeline_updates = []

    # First from any generic components specified
    for component in run_info.get("components", []):
        component_info = components[component]
        param_updates.extend(component_info.get("params", []))
        value_updates.extend(component_info.get("values", []))
        prior_updates.extend(component_info.get("priors", []))
        pipeline_updates.extend(component_info.get("pipeline", []))

    # And then for anything specific to this pipeline
    param_updates.extend(run_info.get("params", []))
    value_updates.extend(run_info.get("values", []))
    prior_updates.extend(run_info.get("priors", []))
    pipeline_updates.extend(run_info.get("pipeline", []))

    # Now apply all the steps
    apply_updates(params, param_updates, is_params=True)
    apply_updates(values, value_updates)
    apply_updates(priors, prior_updates)
    apply_pipeline_updates(params, pipeline_updates)

    output_name = run_info.get("output_name", output_name)
    set_output_dir(params, name, output_dir, output_name)

    # Finally, set the submission information
    submission_info = submission_info.copy()
    submission_info["output_dir"] = output_dir
    submission_info.update(run_info.get("submission", {}))
    run["submission"] = submission_info

    run["params"] = params
    run["values"] = values
    run["priors"] = priors

    return run


def expand_environment_variables(runs):
    """
    For each run, expand any environment variables in the all three parameter files.

    Environment variables are only supported in values, not in keys or sections.

    Parameters
    ----------
    runs : dict
        A dictionary of runs, keyed by name
    """
    for run in runs.values():
        # This sets environment varibles for the duration of the with block
        with temporary_environment(run["env"]):

            # We apply this to all the different ini files
            for ini in [run["params"], run["values"], run["priors"]]:
                for section in ini.sections():
                    # Now we can expand all the actual options
                    for option in ini.options(section):
                        value = ini.get(section, option)
                        new_value = os.path.expandvars(value)
                        if new_value != value:
                            ini.set(section, option, new_value)

def parse_yaml_run_file(run_config):
    """
    Parse a yaml file (or process a previously parsed file)
    containing a list of runs to perform.

    The file should have the following format:

    output_dir: <output directory>

    include: <list of other run files to include>

    components:
        <name of component>:
            params:
            - <list of updates to apply to the parameters file>
            values:
            - <list of updates to apply to the values file>
            priors:
            - <list of updates to apply to the priors file>
            pipeline:
            - <list of updates to apply to the pipeline>
    

    runs:
        - name: <name of the run>
        base: <name of the base parameter file>
        parent: <name of the parent run>
        components:
        - <list of components to include from the list above>
        env:
        <dictionary of environment variables to set>
        params:
        - <list of updates to apply to the parameters file>
        values:
        - <list of updates to apply to the values file>
        priors:
        - <list of updates to apply to the priors file>
        pipeline:
        - <list of updates to apply to the pipeline>

    submission:
        submit: a command to submit runs from batch files (default sbatch)
        cancel: a command to cancel runs from batch files (default scancel)
        template: a template for the batch file (default is a SLURM one suitable for NERSC)
        # remaining variables are passed to the template and can be overridden in specific runs
        time: Wall time for the job, e.g. 00:30:00
        nodes: Number of nodes, e.g. 1
        tasks: Number of tasks in total, e.g. 1
        cores_per_task: Number of cores (threads) per task, e.g. 1
        queue: Queue to submit to, e.g. regular

                
    Parameters
    ----------
    run_config : str or dict
        The name of the file to parse, or a previously loaded configuration

    Returns
    -------
    runs : dict
        A dictionary of runs, keyed by name
    
    components : dict
        A dictionary of components, keyed by name
    """
    if isinstance(run_config, dict):
        info = run_config
    else:
        with open(run_config, 'r') as f:
            info = load_yaml(f)
    
    output_dir = info.get("output_dir", ".")
    output_name = info.get("output_name", "{name}")

    include = info.get("include", [])
    if isinstance(include, str):
        include = [include]

    # Can include another run file, which we deal with
    # recursively.  
    runs = {}
    components = {}
    for include_file in include:
        inc_runs, inc_comps = parse_yaml_run_file(include_file)
        components.update(inc_comps)
        runs.update(inc_runs)

    # But we override the output directory
    # of any imported runs with the one we have here   
    for name, run in runs.items():
        set_output_dir(run["params"], name, output_dir, output_name)
    
    # deal with re-usable components
    components.update(info.get("components", {}))

    submission_info = info.get("submission", {})

    # Build the parameter, value, and prior objects for this run
    for run_dict in info["runs"]:
        name = run_dict["name"]
        runs[name] = build_run(name, run_dict, runs, components, output_dir, submission_info, output_name)

    # Only now do we expand environment variables in the runs.  This gives the child runs
    # a chance to override the environment variables of their parents.
    expand_environment_variables(runs)

    return runs, components


def show_run(run):
    """
    Print a complete description of the run to the screen

    Parameters
    ----------
    run : dict
        The run to print
    
    Returns
    -------
    None
    """
    print(underline(f"Run {run['name']}", '='))
    print(underline("Parameters"))
    run["params"].write(sys.stdout)
    print("")
    print(underline("Values"))
    run["values"].write(sys.stdout)
    print("")
    print(underline("Priors"))
    run["priors"].write(sys.stdout)
    print("")

def perform_test_run(run, use_pdb=False):
    """
    Launch a run under the "test" sampler, which just runs the pipeline
    and does not do any sampling.

    Parameters
    ----------
    run : dict
        The run to perform
    
    Returns
    -------
    status: int
        The exit status of the run
    """
    params = run["params"]
    values = run["values"]
    priors = run["priors"]
    env = run["env"]
    params.set("runtime", "sampler", "test")
    params.set("pipeline", "debug", "T")
    params.set("runtime", "verbosity", "debug")
    params.set("runtime", "resume", "F")

    with temporary_environment(env):
        if use_pdb:
            with run_under_debugger():
                return run_cosmosis(params, values=values, priors=priors)
        else:
            return run_cosmosis(params, values=values, priors=priors)

def chain_status(filename, include_comments=False):
    n = 0
    line = ""
    for line in open(filename):
        if include_comments or not line.startswith("#"):
            n += 1
    # if the last line
    complete = line.startswith("#complete=1")
    last_update_time = os.path.getmtime(filename)
    time_ago_seconds = time.time() - last_update_time
    time_ago_minutes = time_ago_seconds / 60
    return n, complete, time_ago_minutes

def show_run_status(runs, names=None):
    """
    Report the status of the output of one or more runs.

    Parameters
    ----------
    runs : dict
        A dictionary of runs, keyed by name
    names : list
        The names of the runs to report on.  If None then all runs are reported on.
    
    Returns
    -------
    None
    """
    if not names:
        names = runs.keys()

    for name in names:
        run = runs[name]
        sampler = run["params"].get("runtime", "sampler")
        if sampler == "test":
            output_dir = run["params"].get("test", "save_dir")
            if os.path.exists(output_dir):
                print(f"🟢 {name} has been run [test sampler]")
            else:
                print(f"🔴 {name} has not been run [test sampler]")
        else:
            output_file = run["params"].get("output", "filename")
            if os.path.exists(output_file):
                n, complete, last_update = chain_status(output_file)
                if complete:
                    print(f"🟢 {name} output complete with {n} samples, updated {last_update:.1f} minutes ago")
                elif n:
                    print(f"🟡 {name} output exists with {n} samples, updated {last_update:.1f} minutes ago")
                else:
                    print(f"🟠 {name} output exists with 0 samples, updated {last_update:.1f} minutes ago")
            else:
                print(f"🔴 {name} output missing with 0 samples")


def launch_run(run, mpi=False):
    """
    Launch a CosmoSIS run.

    Parameters
    ----------
    run : dict
        The run to launch
    
    Returns
    -------
    status: int
        The exit status of the run
    """
    params = run["params"]
    values = run["values"]
    priors = run["priors"]
    env = run["env"]

    with temporary_environment(env):
        if mpi:
            with MPIPool() as pool:
                return run_cosmosis(params, values=values, priors=priors, pool=pool)
        else:
            return run_cosmosis(params, values=values, priors=priors)


def submit_run(run_file, run):
    """
    Subnmit a CosmoSIS run using a batch system such as SLURM.
    """

    submission_info = run["submission"]
    template = submission_info["template"]
    keys = submission_info.copy()
    name = run["name"]
    keys["job_name"] = name
    output_dir = keys["output_dir"]

    # We use the campaign program to actually run the jobs
    keys["command"]  = f"cosmosis-campaign {run_file} --run {name} --mpi"

    # choose where the jobs stdout / stderr should go
    os.makedirs(os.path.join(output_dir, "logs"), exist_ok=True)
    keys["log"] = os.path.join(output_dir, "logs", f"{name}.log")

    # fill in the template
    sub_script = template.format(**keys).lstrip()

    # write the submission file to a batch subdir
    os.makedirs(os.path.join(output_dir, "batch"), exist_ok=True)
    sub_file = os.path.join(output_dir, "batch", f"{name}.sub")
    with open(sub_file, "w") as f:
        f.write(sub_script)

    # Actually submit the job using slurm or similar
    submit = submission_info.get("submit", "sbatch")
    subprocess.check_call(f"{submit} {sub_file}", shell=True)
    print(f"Submitted {sub_file}\nJob output in", keys["log"])



import argparse
parser = argparse.ArgumentParser(description="Manage and launch CosmoSIS runs")
parser.add_argument("run_config", help="The yaml file containing the runs to perform")
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument("--list", "-l", action="store_true", help="List all available runs")
group.add_argument("--cat", "-c",  type=str, help="Show a single run")
group.add_argument("--status", "-s", default="_unset", nargs="*", help="Show the status of a single run, or all runs if called with no argument")
group.add_argument("--run", "-r",  help="Run the named run")
group.add_argument("--test", "-t",  help="Test the named run")
group.add_argument("--submit", "-x",  help="Submit the named run to a batch system")
parser.add_argument("--mpi", action="store_true", help="Use MPI to launch the runs")
parser.add_argument("--pdb",  action="store_true", help="When testing a run, enter the debugger in case of failure")



def main(args):
    runs, _ = parse_yaml_run_file(args.run_config)

    if args.mpi and not args.run:
        raise ValueError("MPI can only be used when running a single run")

    status_set = args.status != "_unset"

    # The various command line options, which are mutually exclusive.
    # This is enforced by the parser above.
    if args.list:
        for name in runs:
            print(name)
    elif args.cat:
        show_run(runs[args.cat])
    elif args.test:
        perform_test_run(runs[args.test], args.pdb)
    elif status_set:
        show_run_status(runs, args.status)
    elif args.run:
        launch_run(runs[args.run], mpi=args.mpi)
    elif args.submit:
        submit_run(args.run_config, runs[args.submit])



if __name__ == "__main__":
    args = parser.parse_args()
    main(args)




from cobaya.samplers.mcmc.proposal import BlockedProposer

# The cobaya proposal updates the parameter vector in-place,
# so we wrap it slightly to make a copy instead, as expected
# by the metropolis sampler
class CobayaProposalWrapper(BlockedProposer):
    def propose(self, p):
        q = p.copy()
        self.get_proposal(q)
        return q
    def propose_slow(self, p):
        q = p.copy()
        self.get_proposal_slow(q)
        return q
    def propose_fast(self, p):
        q = p.copy()
        self.get_proposal_fast(q)
        return q



import os
import getdist
import numpy as np
from astropy.table import Table

# This isn't used directly, but has the side effect of registering the cosmosis table format:
from . import table as cosmosis_table
from . import postprocessing
from .runtime import Inifile


def cosmosis_to_getdist(filename, name, burn=0, nchain=0):
        """
        Read a cosmosis chain and convert it to a getdist MCSamples object.

        This uses the default latex.ini file bundled with cosmosis to get parameter labels,
        which covers many common parameters, but you might need to modify labels in the
        output object manually if you are using non-covered parameters.

        Parameters
        ----------
        filename : str
            The filename of the cosmosis chain file(s). If nchain>0,
            this should be the root filename without the _1.txt, _2.txt etc suffixes.
            Otherwise it should be the full filename.
        name : str
            The name to assign to the MCSamples object.
        burn : float, optional
            The fraction of samples to discard as burn-in. Default is 0.0
        nchain : int, optional
            The number of chain files to read in. If >0, the function will
            read in multiple files with suffixes _1.txt, _2.txt, etc. Default is 0.

        Returns
        -------
        getdist.MCSamples
            The MCSamples object containing the chain data.
        """
        # Get parameter labels from latex.ini that is bundled with cosmosis
        latex_file = os.path.join(os.path.dirname(postprocessing.__file__), "latex.ini")
        latex_names = Inifile(latex_file)

        # If nchain is specified, read in multiple chain files - this is usually
        # for the metropolis sampler under MPI.
        if nchain > 0:
            filenames = [filename + f"_{i}.txt" for i in range(1, nchain+1)]
            samples = []
            weights = []
            loglikes = []

            # MCSamples can accept a list of arrays when multiple chains
            # are generated, so read each chain file in turn and append
            # the data to lists.
            for filename in filenames:
                table = Table.read(filename, format='cosmosis')

                # Read the chain data for this file.
                # The metadata should be the same for all chains.
                samples_i, weights_i, names, loglikes_i, labels, sampler_type = _extract_getdist_inputs(
                    table, latex_names)
                samples.append(samples_i)
                weights.append(weights_i)
                loglikes.append(loglikes_i)
        else:
            # Otherwise if there is just a single chain file, read that in,
            # as MCSamples can also be given single arrays.
            t = Table.read(filename, format='cosmosis')
            samples, weights, names, loglikes, labels, sampler_type = _extract_getdist_inputs(t, latex_names)

        # Create and return the MCSamples object that you can use with getdist
        # as normal
        s = getdist.MCSamples(
                        samples=samples,
                        weights=weights,
                        names=names,
                        name_tag=name,
                        loglikes=loglikes,
                        labels=labels,
                        ignore_rows=burn,
                        sampler=sampler_type,
                    )

        return s




def _extract_getdist_inputs(table, latex_names):
    # Many cosmosis chains are weighted; get the weights here
    if "weight" in table.colnames:
        weights = table["weight"]
    elif "log_weight" in table.colnames:
        weights = np.exp(table["log_weight"] - np.max(table["log_weight"]))
    else:
        weights = None

    # Getdist calls the -log(posterior) column "loglikes"
    # Extract that column here if it exists, which it should in all
    # normal circumstances
    if "post" in table.colnames:
        loglikes = -table["post"]
    else:
        loglikes = None
        
    # Get the samples themselves as an array,
    # excluding the cosmosis special columns
    # Also get the parameter names.
    excluded = ["weight", "log_weight", "prior", "like", "post"]
    samples = np.array([table[col] for col in table.colnames if col not in excluded]).T
    names = [col for col in table.colnames if col not in excluded]

    # If this is an emcee chain, we need to reshape the samples
    # so that each walker is a separate chain.
    # emcee does not have a weight column so we don't have to 
    # change the shape of that.
    if table.meta["sampler"] == "emcee":
        samples = [samples[i::table.meta["walkers"]] for i in range(table.meta["walkers"])]
        loglikes = [loglikes[i::table.meta["walkers"]] for i in range(table.meta["walkers"])]

    # Get the sample labels from the latex.
    labels = names[:]
    for i, name in  enumerate(names):
        if "--" in name:
            sec, key = name.lower().split("--")
            if latex_names.has_option(sec, key):
                labels[i] = latex_names[sec, key]

    # The nested samplers in cosmosis store the logZ in the final metadata
    # (i.e. the metdata generated at the end of the run and stored at the 
    # end of the file). If this exists, we can assume this is a nested
    # sampler run, otherwise it's MCMC.
    if "final_metadata_items" in table.meta:
        if "log_z" in table.meta["final_metadata_items"]:
            sampler_type = "nested"
        else:
            sampler_type = "mcmc"

    return samples, weights, names, loglikes, labels, sampler_type



from ..campaign import *
from ..runtime import Inifile
import os
import tempfile
import contextlib
import pytest

NRUN = 15

@contextlib.contextmanager
def run_from_source_dir():
    this_dir = os.path.split(os.path.abspath(__file__))[0]
    source_dir = os.path.join(this_dir, "..", "..")
    cur_dir = os.getcwd()
    try:
        os.chdir(source_dir)
        yield
    finally:
        os.chdir(cur_dir)


def test_pipeline_after():
    params = Inifile(None)
    params.add_section("pipeline")
    params.set("pipeline", "modules", "a b c")
    pipeline_after(params, "a", "d")
    assert params.get("pipeline", "modules") == "a d b c"
    pipeline_after(params, "c", ["e", "f"])
    assert params.get("pipeline", "modules") == "a d b c e f"

def test_pipeline_before():
    params = Inifile(None)
    params.add_section("pipeline")
    params.set("pipeline", "modules", "a b c")
    pipeline_before(params, "a", "d")
    assert params.get("pipeline", "modules") == "d a b c"
    pipeline_before(params, "c", ["e", "f"])
    assert params.get("pipeline", "modules") == "d a b e f c"

def test_pipeline_replace():
    params = Inifile(None)
    params.add_section("pipeline")
    params.set("pipeline", "modules", "a b c")
    pipeline_replace(params, "b", "d")
    assert params.get("pipeline", "modules") == "a d c"
    pipeline_replace(params, "c", ["e", "f"])
    assert params.get("pipeline", "modules") == "a d e f"

def test_pipeline_delete():
    params = Inifile(None)
    params.add_section("pipeline")
    params.set("pipeline", "modules", "a b c")
    pipeline_delete(params, "b")
    assert params.get("pipeline", "modules") == "a c"
    pipeline_delete(params, ["a", "c"])
    assert params.get("pipeline", "modules") == ""

def test_pipeline_append():
    params = Inifile(None)
    params.add_section("pipeline")
    params.set("pipeline", "modules", "a b c")
    pipeline_append(params, ["d"])
    assert params["pipeline", "modules"] == "a b c d"

def test_pipeline_prepend():
    params = Inifile(None)
    params.add_section("pipeline")
    params.set("pipeline", "modules", "a b c")
    pipeline_prepend(params, ["d", "e"])
    assert params["pipeline", "modules"] == "d e a b c"


def test_campaign_functions():
    with run_from_source_dir():
        runs, _ = parse_yaml_run_file("cosmosis/test/campaign.yml")

        assert len(runs) == NRUN
        assert "v1" in runs
        assert runs["v2"]["values"].get("parameters", "p1") == "-2.0 0.0 2.0"
        assert runs["v2"]["priors"].get("parameters", "p2") == "gaussian 0.0 1.0"
        assert runs["v3"]["params"].get("runtime", "sampler") == "emcee"
        assert runs["v3"]["params"].get("emcee", "walkers") == "8"

        assert not runs["v4"]["priors"].has_option("parameters", "p2")

        assert runs["v3"]["params"].get("output", "filename") == "output/campaign-test/my_project_v3_suite1.txt"

        for name in runs:
            print(name)

        show_run(runs["v1"])
        perform_test_run(runs["v1"])
        perform_test_run(runs["v1"], use_pdb=True)
        show_run_status(runs)
        show_run_status(runs, ["v1"])
        show_run_status(runs, ["v2"])
        perform_test_run(runs["v2"])
        show_run_status(runs, ["v1"])
        show_run_status(runs, ["v2"])
        launch_run(runs["v2"])
        show_run_status(runs, ["v1"])
        show_run_status(runs, ["v2"])

        submit_run("cosmosis/test/campaign.yml", runs["v3"])
        submit_run("cosmosis/test/campaign.yml", runs["v4"])

def test_campaign_functions2():
    with run_from_source_dir():
        with open("cosmosis/test/campaign.yml") as f:
            runs_config = load_yaml(f)

    with tempfile.TemporaryDirectory() as dirname:
        runs_config['output_dir']  = dirname

        with run_from_source_dir():
            with open("cosmosis/test/campaign.yml") as f:
                runs, _ = parse_yaml_run_file(runs_config)

            for name in runs:
                print(name)


            assert len(runs) == NRUN
            assert "v1" in runs
            assert runs["v2"]["values"].get("parameters", "p1") == "-2.0 0.0 2.0"
            assert runs["v2"]["priors"].get("parameters", "p2") == "gaussian 0.0 1.0"
            assert runs["v3"]["params"].get("runtime", "sampler") == "emcee"
            assert runs["v3"]["params"].get("emcee", "walkers") == "8"

            assert not runs["v4"]["priors"].has_option("parameters", "p2")

            show_run(runs["v1"])
            perform_test_run(runs["v1"])
            expected_test_output_dir = os.path.join(dirname, "my_project_v1_suite1")
            assert os.path.isdir(expected_test_output_dir)
            show_run_status(runs)
            show_run_status(runs, ["v1"])
            show_run_status(runs, ["v2"])
            perform_test_run(runs["v2"])
            show_run_status(runs, ["v1"])
            show_run_status(runs, ["v2"])
            launch_run(runs["v2"])
            show_run_status(runs, ["v1"])
            show_run_status(runs, ["v2"])

            submit_run("cosmosis/test/campaign.yml", runs["v3"])
            submit_run("cosmosis/test/campaign.yml", runs["v4"])


def test_polychord_multinest_campaign():
    with run_from_source_dir():
        with open("cosmosis/test/campaign.yml") as f:
            runs_config = load_yaml(f)

    with tempfile.TemporaryDirectory() as dirname:
        runs_config['output_dir']  = dirname

        with run_from_source_dir():
            with open("cosmosis/test/campaign.yml") as f:
                runs, _ = parse_yaml_run_file(runs_config)

            for name in runs:
                print(name)

            # polychord and multinest tests - should make the extra output files
            launch_run(runs["multinest-test"])
            launch_run(runs["polychord-test"])
            mn_file = os.path.join(dirname, "my_project_multinest-test_suite1.multinest.txt")
            pc_file = os.path.join(dirname, "my_project_polychord-test_suite1.polychord.txt")
            assert os.path.isfile(mn_file)
            assert os.path.isfile(pc_file)



def test_campaign_env():
    os.environ["TEST"] = "aaa"
    with run_from_source_dir():
        runs, _ = parse_yaml_run_file("cosmosis/test/campaign.yml")
    assert runs["env-test-1"]["params"].get("test1", "env_test_var")  == "xxx"
    assert runs["env-test-2"]["params"].get("test1", "env_test_var")  == "yyy"

def test_inherit_env():
    with run_from_source_dir():
        runs, _ = parse_yaml_run_file("cosmosis/test/campaign.yml")
    assert runs["env-test-3"]["params"].get("test1", "env_test_var")  == "xxx"
    assert runs["env-test-3"]["params"].get("emcee", "walkers")  == "xxx"
    assert runs["env-test-5"]["params"].get("test1", "env_test_var")  == "yyy"
    assert runs["env-test-5"]["params"].get("test1", "env_test_var2")  == "zzz"


def test_campaign_duplicate_keys():
    with pytest.raises(ValueError):
        with run_from_source_dir():
            runs, _ = parse_yaml_run_file("cosmosis/test/bad-campaign.yml")

def test_component():
    with run_from_source_dir():
        runs, components = parse_yaml_run_file("cosmosis/test/campaign.yml")
    assert "test_component_1" in components
    assert runs["component-test"]["params"].get("emcee", "walkers") == "100"

def test_include():
    with run_from_source_dir():
        runs, _ = parse_yaml_run_file("cosmosis/test/campaign.yml")
    assert "imported-run" in runs
    assert "include-test-1" in runs
    assert runs['include-test-1']['params'].get("emcee", "walkers") == "755"
    assert "include-test-2" in runs
    assert runs['include-test-2']['params'].get("emcee", "walkers") == "200"



"""
Import this module to register the cosmosis table format with astropy.

It's not quite a perfect round trip yet, but it's close enough for almost
any purpose.
"""
from cosmosis.output import utils
from astropy.table import Table
from astropy.io import registry
import os
from typing import Union, IO

def openr(f):
    if isinstance(f, str):
        return open(f, 'r')
    # Doesn't have to be file-like - can be a list of lines
    return f

def openw(f):
    if isinstance(f, str):
        return open(f, 'w')
    elif hasattr(f, 'write'):
        return f
    else:
        raise ValueError("Argument must be a filename or a file-like object")

def read_cosmosis_table(input_file: Union[str, IO, list], delimiter: str=None) -> Table:
    """
    Read a cosmosis-format text file into an astropy Table.

    This will store the metadata in the Table.meta dictionary.
    Metadata that appears after the data is stored with a special key
    "final_metadata_items" which is a list of the keys that were in
    the final metadata.

    Parameters
    ----------
    input_file : str or file-like or list
        The name of the file to read, or an open file-like object,
        or a list of lines (e.g. from f.readlines()).
    delimiter : str, optional
        The delimiter used in the file. If None (default) then
        a tab is used.

    Returns
    -------
    table : astropy.table.Table
    """
    chain = []
    metadata = {}
    final_metadata = {}
    comments = []
    started_data = False

    # This text is copied from the text_output.py load_from_options method.
    # On the major TODO list for cosmosis v4 is to refactor the I/O more generally,
    # so I won't try to avoid the copy/paste for now.
    for i, line in enumerate(openr(input_file)):
        line = line.strip()
        if not line:
            continue

        if line.startswith('#'):
            # remove the first #
            # if there is another then this is a comment,
            # not metadata
            line = line[1:]
            if i == 0:
                column_names = line.split()
            elif line.startswith('#'):
                comment = line[1:].lstrip()
                comments.append(comment)
            else:
                #parse form '#key=value #comment'
                if line.count('#') == 0:
                    key_val = line.strip()
                    comment = ''
                else:
                    key_val, comment = line.split('#', 1)
                key,val = key_val.split('=',1)
                val = utils.parse_value(val)

                if started_data:
                    final_metadata[key] = val
                else:
                    metadata[key] = val
        else:
            started_data = True
            words = line.split(delimiter)
            vals = [float(word) for word in words]
            chain.append(vals)

    data = Table(rows=chain, names=column_names)

    # Handle metadata. CosmoSIS distinguishes between metadata
    # that comes before the data and after the data. We store
    # both, and also store a list of which keys were in the
    # final metadata, so that writers can put them at the end
    # of the file if they want to.
    data.meta.update(metadata)
    data.meta.update(final_metadata)
    data.meta["final_metadata_items"] = list(final_metadata.keys())

    if comments:
        data.meta['comments'] = comments

    return data


def write_cosmosis_table(table: Table, filename: Union[str, IO], overwrite: bool=False, delimiter: str="\t"):
    """
    Write an astropy Table to a cosmosis-format text file.

    This will write the metadata in the Table.meta dictionary.
    Metadata that appears after the data should be listed in
    the "final_metadata_items" key in the meta dictionary.

    Parameters
    ----------
    table : astropy.table.Table
        The table to write.

    filename : str or file-like
        The name of the file to write, or an open file-like object.
    
    overwrite : bool, optional
        If True, overwrite an existing file. Default is False,
        which will raise an error if the file exists.
    
    delimiter : str, optional
        The delimiter to use in the file. Default is tab.
    """
    if isinstance(filename, str) and (not overwrite) and os.path.exists(filename):
        raise OSError(f"File {filename} exists and overwrite=False")

    with openw(filename) as f:

        # Write the main header line.
        f.write("#" + delimiter.join(table.colnames) + "\n")

        # We will defer writing the final metadata until the end.
        final_items = table.meta.get("final_metadata_items", [])

        # Write main metadata first at the top
        for key, val in table.meta.items():
            if key == "comments" or key == "final_metadata_items":
                continue

            if key in final_items:
                continue
                
            f.write(f"#{key}={val}\n")

        # Now write any comment metadata, which is saved as
        # a double-hashed comment line.
        comments = table.meta.get("comments", [])
        for line in comments:
            f.write(f"## {line}\n")

        # Write the data itself
        for row in table:
            line = delimiter.join(str(x) for x in row) + '\n'
            f.write(line)

        # Write the final metadata at the end of the file
        for key in final_items:
            val = table.meta[key]
            f.write(f"#{key}={val}\n")

# Tell astropy about this format, so that you can write format="cosmosis"
# in astropy table read and write calls.
registry.register_reader('cosmosis', Table, read_cosmosis_table)
registry.register_writer('cosmosis', Table, write_cosmosis_table)




#coding: utf-8

u"""Definition of the :class:`DataBlock` class."""

import ctypes as ct
from . import errors
from . import dbt_types as types
from ...utils import mkdir
from .errors import BlockError
import numpy as np
import os
import collections
import tarfile
import io
from io import StringIO, BytesIO
import sys



try:
	from . import lib
except:
	_allow_unbuilt_import = os.environ.get("COSMOSIS_ALLOW_UNBUILT_IMPORT", "0") == "1"
	if _allow_unbuilt_import:
		print("Warning: importing cosmosis datablock without compiled library. "
			  "You cannot use CosmoSIS properly; this is designed for building "
			  "documentation", file=sys.stderr)
	else:
		raise



option_section = "module_options"
metadata_prefix = "cosmosis_metadata:"

class DataBlock(object):
	u"""A map of (section,name)->value of parameters.

	At the heart of Cosmosis is a data-containing object which is passed
	down a pipeline of processing stages, which shape and massage those
	data as they go through.  The :class:`DataBlock` class is the
	realization of this object as seen by Python modules.

	The main methods a Cosmosis module programmer is interested in given
	one of these objects are the implicitly-called `__getitem__` and
	`__setitem__`: these retrieve parameter values from the map, and put
	new ones in or replace existing ones, respectively.

	Most of the implementation detail of this class is a complete
	orthogonal set of methods which get, put and replace parameters with
	integer, boolean, string, floating-point, complex values, either as
	scalars or 1-, 2-dimensional arrays or ‘grids’, then refinement of
	these into generic :func:`get`, :func:`set` and :func:`replace`
	methods, and finally the ultimate refinement to the
	:func:`__getitem__` and :func:`__setitem__` methods themselves.

	The *grid* concept is where a two-dimensional array is flanked by two
	one-dimensional ones giving labels to the ‘rows’ and ‘columns’; these
	labels are used to address the data directly.

	"""

	GET=0
	PUT=1
	REPLACE=2
	def __init__(self, ptr=None, own=None):
		u"""Construct an empty parameter map, or possibly shadow an existing one.

		In implementation, this Python object is actually a wrapper around
		a C object.  The constructor allows for an existing object to be
		specified through the `ptr`, and then to dictate that the Python
		object is ultimately responsible for the lifetime of the
		underlying object, via the boolean-valued `own`.

		Note that it is also possible to not specify `ptr` and to specify
		`own` as `False`, in which case a new C object will be created but
		it will be left to the application to ensure proper destruction at
		the end of its lifetime.

		"""

		# Doc: Need to find out the use-case for this latter option,
		#      and if it is being used in Cosmosis now; if not, it
		#      should be removed!

		self.owns=own
		if ptr is None:
			ptr = lib.make_c_datablock()
			self.owns=True
		if own is not None:
			self.owns=own
		self._ptr = ptr
		self._as_parameter_ = ptr
	#TODO: add destructor.  destroy block if owned

	def __del__(self):
		u"""Destroy this object.

		Also destroy the underlying C object if we are deemed to own it.

		"""
		try:
			if self.owns:
				lib.destroy_c_datablock(self._ptr)
		except:
			pass

				
	def clone(self):
		u"""Make a brand-new, completely independent object, a deep copy of the existing one.

		A new object will be returned from this method which has its own
		underlying implementation, a deep copy of the parameter map we
		are holding.  This WILL entail the attempted requisition of
		enough new memory to hold the complete parameter structure.

		"""
		ptr = lib.clone_c_datablock(self._ptr)
		return DataBlock(ptr,own=True)


	@staticmethod
	def python_to_c_complex(value):
		u"""Interpret an arbitrary Python object as a lib.c_complex type.

		This convenience function will take an actual lib.c_complex
		`value` (no-op), a Python complex `value`, the first two
		components of a Python tuple `value`, or a real scalar `value` and
		return the equivalent lib.c_complex (i.e. a type which can be
		passed to a C subroutine representing a complex number).

		In the case of the scalar input, this is taken as the real part of
		the complex number and the imaginary part will be zero.

		"""
		if isinstance(value, lib.c_complex):
			return value
		elif isinstance(value, complex):
			return lib.c_complex(value.real,value.imag)
		elif isinstance(value, tuple):
			return lib.c_complex(value[0],value[1])
		else:
			return lib.c_complex(value, 0.0)

	@staticmethod
	def python_to_1d_c_array(value, numpy_type):
		u"""Create a C object equivalent to the `value` array, interpreted as `numpy_type`.

		The object will be a contiguous list—this may entail that a value
		array with strides be copied to a compressed version—of C type
		most appropriate to the representation of the Python `numpy_type`.

		"""
		value = np.array(value, dtype=numpy_type)
		#This function is for 1D arrays only
		assert value.ndim==1
		#check strides same as itemsize.
		#This may be false if e.g. the object
		#was made by looking at a slice through
		#a matrix.
		if value.itemsize != value.strides[0]:
			#If so we need to copy the data to create
			#a new object with sensible strides
			value = value.copy()
		assert value.itemsize==value.strides[0]
		#Now return pointer to start of the data
		array = np.ctypeslib.as_ctypes(value)
		array_size = value.size
		#OK, here's the difficult part.
		# We have to return the value, as well as the
		# array we have converted it to.
		# That's because the array object
		# does not maintain a pointer to the value,
		# so it is garbage collected if we don't
		# and then the array just contains junk memory
		return value, array, array_size



	def get_int(self, section, name, default=None):
		u"""Retrieve an integer value from the parameter set.

		The `name` ʼd parameter in the given `section` will be interpreted
		as an integer and returned to the caller.  If such parameter is
		not found in the map, then the `default` will be returned if it
		was given, or else a specialized :class:`BlockError` (see
		errors.py) will be thrown.  The :class:`BlockError` may also be
		thrown if a variable is found, but is not of integer type.

		"""
		r = ct.c_int()
		if default is None:
			status = lib.c_datablock_get_int(self._ptr,section.encode('ascii'),name.encode('ascii'),r)
		else:
			status = lib.c_datablock_get_int_default(self._ptr,section.encode('ascii'),name.encode('ascii'),default,r)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)
		return r.value

	def get_bool(self, section, name, default=None):
		u"""Retrieve a boolean value from the parameter set.

		The `name` parameter in the given `section` will be interpreted
		as a boolean and returned to the caller.  If such parameter is not
		found in the map, then the `default` will be returned if it was
		given, or else a specialized :class:`BlockError` (see errors.py)
		will be thrown.  The :class:`BlockError` may also be thrown if a
		variable is found, but is not of boolean type.

		"""
		r = ct.c_bool()
		if default is None:
			status = lib.c_datablock_get_bool(self._ptr,section.encode('ascii'),name.encode('ascii'),r)
		else:
			status = lib.c_datablock_get_bool_default(self._ptr,section.encode('ascii'),name.encode('ascii'),default,r)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)
		return r.value

	def get_double(self, section, name, default=None):
		u"""Retrieve a floating-point value from the parameter set.

		The `name` parameter in the given `section` will be interpreted
		as a floating-point value and returned to the caller.  If such
		parameter is not found in the map, then the `default` will be
		returned if it was given, or else a specialized
		:class:`BlockError` (see errors.py) will be thrown.  The
		:class:`BlockError` may also be thrown if a variable is found, but
		is not of floating-point type.

		"""
		r = ct.c_double()
		if default is None:
			status = lib.c_datablock_get_double(self._ptr,section.encode('ascii'),name.encode('ascii'),r)
		else:
			status = lib.c_datablock_get_double_default(self._ptr,section.encode('ascii'),name.encode('ascii'),default,r)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)
		return r.value

	def get_complex(self, section, name, default=None):
		u"""Retrieve a complex value from the parameter set.

		The `name` parameter in the given `section` will be interpreted
		as a complex value and returned to the caller.  If such parameter
		is not found in the map, then the `default` will be returned if it
		was given, or else a specialized :class:`BlockError` (see
		errors.py) will be thrown.  The :class:`BlockError` may also be
		thrown if a variable is found, but is not of complex type.

		"""
		r = lib.c_complex()
		if default is None:
			status = lib.c_datablock_get_complex(self._ptr,section.encode('ascii'),name.encode('ascii'),r)
		else:
			status = lib.c_datablock_get_complex_default(self._ptr,section.encode('ascii'),name.encode('ascii'),default,r)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)
		return r.real+1j*r.imag

	def get_string(self, section, name, default=None):
		u"""Retrieve a string value from the parameter set.

		The `name` parameter in the given `section` will be interpreted
		as a string value and returned to the caller.  If such parameter
		is not found in the map, then the `default` will be returned if it
		was given, or else a specialized :class:`BlockError` (see
		errors.py) will be thrown.  The :class:`BlockError` may also be
		thrown if a variable is found, but is not of string type.

		"""
		r = lib.c_str()
		if default is None:
			status = lib.c_datablock_get_string(self._ptr,section.encode('ascii'),name.encode('ascii'),r)
		else:
			status = lib.c_datablock_get_string_default(self._ptr,section.encode('ascii'),name.encode('ascii'),default.encode('ascii'),r)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)
		c = r.value.decode('utf-8')
		lib.free(r)
		return c

	def get_int_array_1d(self, section, name):
		u"""Retrieve an integer array from the parameter set.

		The `name` parameter in the given `section` will be understood
		as being of integer array type and returned to the caller as a
		NumPy array.  If such a parameter is not found in the map, then a
		specialized :class:`BlockError` (see errors.py) will be thrown.

		"""
		n = lib.c_datablock_get_array_length(self._ptr, section.encode('ascii'), name.encode('ascii'))
		r = np.zeros(n, dtype=np.intc)
		arr = np.ctypeslib.as_ctypes(r)
		sz = lib.c_int()
		status = lib.c_datablock_get_int_array_1d_preallocated(self._ptr, section.encode('ascii'), name.encode('ascii'), arr, ct.byref(sz), n)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)
		return r

	def get_double_array_1d(self, section, name):
		u"""Retrieve a floating-point array from the parameter set.

		The `name` parameter in the given `section` will be understood
		as being of floating-point array type and returned to the caller
		as a *NumPy* array.  If such a parameter is not found in the map,
		then a specialized :class:`BlockError` (see errors.py) will be
		thrown.

		"""
		n = lib.c_datablock_get_array_length(self._ptr, section.encode('ascii'), name.encode('ascii'))
		r = np.zeros(n, dtype=np.double)
		arr = np.ctypeslib.as_ctypes(r)
		sz = lib.c_int()
		status = lib.c_datablock_get_double_array_1d_preallocated(self._ptr, section.encode('ascii'), name.encode('ascii'), arr, ct.byref(sz), n)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)
		return r

	def get_string_array_1d(self, section, name):
		u"""Retrieve an array of strings from the datablock.

		The `name` parameter in the given `section` will be understood
		as being of 1D string array type and returned to the caller
		as a numpy array.  If such a parameter is not found in the map,
		then a specialized :class:`BlockError` (see errors.py) will be
		thrown.

		"""
		section = section.encode('ascii')
		name = name.encode('ascii')
		n = lib.c_datablock_get_array_length(self._ptr, section, name)

		# Create the array, but make it of void pointers.
		# If you use c_char_p it just comes through as an int
		# which we can free later
		array = (ct.c_void_p * n)()
		sz = lib.c_int()
		status = lib.c_datablock_get_str_array_1d_preallocated(self._ptr, section, name, array, ct.byref(sz))
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

		# Parse each element into a list.  We don't yet know that max
		# length of the strings so don't want to make a numpy array yet.
		output = []
		for i in range(n):
			# Convert to a char pointer
			c = ct.c_char_p(array[i])
			# Decode char pointer as strings
			output.append(c.value.decode('ascii'))
			# Free the pointer allocated by strdup in C
			lib.free(array[i])

		# Convert to array for output
		return np.array(output)



	def _get_array_nd(self, section, name, dtype):

		if dtype is complex or dtype is str:
			raise ValueError("Sorry - cosmosis support for 2D complex and string values is incomplete")

		ndim = lib.c_int()
		status = lib.c_datablock_get_array_ndim(self._ptr, section.encode('ascii'), name.encode('ascii'), ct.byref(ndim))
		if status:
			raise BlockError.exception_for_status(status, section, name)

		ctype, shape_function, get_function = {
			int: (ct.c_int, lib.c_datablock_get_int_array_shape, lib.c_datablock_get_int_array),
			float: (ct.c_double, lib.c_datablock_get_double_array_shape, lib.c_datablock_get_double_array),
			#complex: (lib.c_complex, lib.c_datablock_get_complex_array_shape, lib.c_datablock_get_complex_array),
			#str: (lib.c_str, lib.c_datablock_get_string_array_shape, lib.c_datablock_get_string_array),
		}[dtype]

		#Get the array extent
		extent = (ct.c_int * ndim.value)()
		status = shape_function(self._ptr, section.encode('ascii'), name.encode('ascii'), ndim, extent)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

		#Make the space for it
		N = tuple([extent[i] for i in range(ndim.value)])
		r = np.zeros(N, dtype=ctype)
		arr = r.ctypes.data_as(ct.POINTER(ctype))

		#Fill in with the data
		status = get_function(self._ptr, section.encode('ascii'), name.encode('ascii'), arr, ndim, extent)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)
		return r

	def _put_replace_array_nd(self, section, name, value, dtype, mode):
		shape = value.shape
		ndim = len(shape)
		extent = (ct.c_int * ndim)()
		for i in range(ndim): extent[i] = shape[i]
		value = value.flatten()
		p, arr, arr_size = self.python_to_1d_c_array(value, dtype)
		put_function={
			(np.intc, self.PUT):lib.c_datablock_put_int_array,
			(np.double, self.PUT):lib.c_datablock_put_double_array,
			(np.intc, self.REPLACE):lib.c_datablock_replace_int_array,
			(np.double, self.REPLACE):lib.c_datablock_replace_double_array,
		}.get((dtype,mode))
		if put_function is None:
			raise ValueError("I do not know how to save %s in %s of type %s"% (section,name, dtype))
		status = put_function(self._ptr, section.encode('ascii'), name.encode('ascii'), arr, ndim, extent)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)


	def put_double_array_nd(self, section, name, value):
		u"""Add a floating-point array parameter to the data set.

		The `value` must be an array of values which can be interpreted as
		floating-point numbers, otherwise a :class:`ValueError` will be
		raised.  If the parameter does not exist in the data set, a
		:class:`BlockError` will be raised.  The array can be any shape.

		"""
		self._put_replace_array_nd(section, name, value, np.double, self.PUT)

	def put_int_array_nd(self, section, name, value):
		u"""Add an integer array parameter to the data set.

		The value must be an array of values which can be interpreted as
		integer numbers, otherwise a :class:`ValueError` will be raised.
		If the parameter does not exist in the data set, a
		:class:`BlockError` will be raised.  The array can be any shape.

		"""
		self._put_replace_array_nd(section, name, value, np.intc, self.PUT)



	def replace_double_array_nd(self, section, name, value):
		u"""Replace a floating-point array parameter in the data set.

		The value must be an array of values which can be interpreted as
		floating-point numbers, otherwise a :class:`ValueError` will be
		raised.  If the parameter already exists in the data set, a
		:class:`BlockError` will be raised.  The new array can be any
		shape, independent of the shape of the original value in this data
		set.

		"""
		self._put_replace_array_nd(section, name, value, np.double, self.REPLACE)

	def replace_int_array_nd(self, section, name, value):
		u"""Replace an integer array parameter in the data set.

		The value must be an array of values which can be interpreted as
		integer numbers, otherwise a :class:`ValueError` will be raised.
		If the parameter already exists in the data set, a
		:class:`BlockError` will be raised.  The new array can be any
		shape, independent of the shape of the original value in this data
		set.

		"""
		self._put_replace_array_nd(section, name, value, np.intc, self.REPLACE)

	def get_double_array_nd(self, section, name):
		u"""Get a floating-point array of *a priori* unspecified shape.

		Expect :class:`BlockError` or :class:`ValueError` to be raised if
		there are extenuating circumstances.

		"""
		return self._get_array_nd(section, name, float)

	def get_int_array_nd(self, section, name):
		u"""Get an integer-valued array of *a priori* unspecified shape.

		Expect :class:`BlockError` or :class:`ValueError` to be raised if
		there are extenuating circumstances.

		"""
		return self._get_array_nd(section, name, int)

	#def get_complex_array_2d(self, section, name):
	#	return self._get_array_2d(section, name, complex)

	#def get_string_array_2d(self, section, name):
	#	return self._get_array_2d(section, name, str)

	def put_int(self, section, name, value):
		u"""Add an integer parameter to the map.

		A new parameter will be added to the current map, at (`section`,
		`name`), and will have the `value` interpreted as an integer type.
		It is an error to try to add a parameter which is already there,
		and in this case a specialized :class:`BlockError` will be raised.

		"""
		status = lib.c_datablock_put_int(self._ptr,section.encode('ascii'),name.encode('ascii'),int(value))
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def put_bool(self, section, name, value):
		u"""Add a boolean parameter to the map.

		A new parameter will be added to the current map, at (`section`,
		`name`), and will have the `value` interpreted as a boolean type.
		It is an error to try to add a parameter which is already there,
		and in this case a :class:`BlockError` will be raised.

		"""
		status = lib.c_datablock_put_bool(self._ptr,section.encode('ascii'),name.encode('ascii'),bool(value))
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def put_double(self, section, name, value):
		u"""Add a floating-point parameter to the map.

		A new parameter will be added to the current map, at (`section`,
		`name`), and will have the `value` interpreted as a floating-point
		type.  It is an error to try to add a parameter which is already
		there, and in this case a :class:`BlockError` will be raised.

		"""
		status = lib.c_datablock_put_double(self._ptr,section.encode('ascii'),name.encode('ascii'),float(value))
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def put_complex(self, section, name, value):
		u"""Add a complex parameter to the map.

		A new parameter will be added to the current map, at (`section`,
		`name`), and will have the `value` interpreted as a complex type.
		It is an error to try to add a parameter which is already there,
		and in this case a :class:`BlockError` will be raised.

		"""
		value=self.python_to_c_complex(value)
		status = lib.c_datablock_put_complex(self._ptr,section.encode('ascii'),name.encode('ascii'),value)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def put_string(self, section, name, value):
		u"""Add a string parameter to the map.

		A new parameter will be added to the current map, at (`section`,
		`name`), and will have the `value` interpreted as a string type.
		It is an error to try to add a parameter which is already there,
		and in this case a :class:`BlockError` will be raised.

		"""
		status = lib.c_datablock_put_string(self._ptr,section.encode('ascii'),name.encode('ascii'),str(value).encode('ascii'))
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def put_int_array_1d(self, section, name, value):
		u"""Add a one-dimensional integer array to the map.

		A parameter called `name` is added to `section`, and holds `value`
		interpreted as a simple array of integers.  If this interpretation
		cannot be made then a :class:`BlockError` will be raised.

		"""
		value_ref, value,n=self.python_to_1d_c_array(value, np.intc)
		status = lib.c_datablock_put_int_array_1d(self._ptr, section.encode('ascii'), name.encode('ascii'), value, n)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def put_double_array_1d(self, section, name, value):
		u"""Add a one-dimensional floating-point array to the map.

		A parameter called `name` is added to `section`, and holds `value`
		interpreted as a simple array of floating-point values.  If this
		interpretation cannot be made then a :class:`BlockError` will be
		raised.

		"""
		value_ref, value,n=self.python_to_1d_c_array(value, np.double)
		status = lib.c_datablock_put_double_array_1d(self._ptr, section.encode('ascii'), name.encode('ascii'), value, n)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def put_string_array_1d(self, section, name, value):
		u"""Add a one-dimensional floating-point array to the map.

		A parameter called `name` is added to `section`, and holds `value`
		interpreted as a simple array of floating-point values.  If this
		interpretation cannot be made then a :class:`BlockError` will be
		raised.

		"""
		array = (ct.c_char_p * len(value))()
		for i, v in enumerate(value):
			array[i] = v.encode('ascii')

		status = lib.c_datablock_put_str_array_1d(self._ptr, section.encode('ascii'), name.encode('ascii'), array, len(value))

		if status!=0:
			raise BlockError.exception_for_status(status, section, name)




	def _method_for_type(self, T, method_type):
		method={ int:    (self.get_int,     self.put_int,     self.replace_int),
				 float:  (self.get_double,  self.put_double,  self.replace_double),
				 bool:   (self.get_bool,    self.put_bool,    self.replace_bool),
				 complex:(self.get_complex, self.put_complex, self.replace_complex),
				 str:    (self.get_string,  self.put_string,  self.replace_string)
				 }.get(T)
		if method:
			return method[method_type]
		return None

	def _method_for_datatype_code(self, code, method_type):

		T={ 
			types.DBT_INT:     (self.get_int,     self.put_int,     self.replace_int),
			types.DBT_BOOL:    (self.get_bool,     self.put_bool,     self.replace_bool),
			types.DBT_DOUBLE:         (self.get_double,  self.put_double,  self.replace_double),
			types.DBT_COMPLEX: (self.get_complex, self.put_complex, self.replace_complex),
			types.DBT_STRING:  (self.get_string,  self.put_string,  self.replace_string),
			types.DBT_INT1D:   (self.get_int_array_1d,     self.put_int_array_1d,     self.replace_int_array_1d),
			types.DBT_DOUBLE1D:(self.get_double_array_1d,  self.put_double_array_1d,  self.replace_double_array_1d),
			# types.COMPLEX1D:   (self.get_complex_array_1d, self.put_complex_array_1d, self.replace_complex_array_1d),
			types.DBT_STRING1D:    (self.get_string_array_1d,  self.put_string_array_1d,  self.replace_string_array_1d),
			# types.DBT_INT2D:   (self.get_int_array_2d,     self.put_int_array_2d,     self.replace_int_array_2d)
			types.DBT_DOUBLEND:(self.get_double_array_nd,  self.put_double_array_nd, None),
			types.DBT_INTND:(self.get_int_array_nd,  self.put_int_array_nd, None),
			# types.COMPLEX2D:   (self.get_complex_array_2d, self.put_complex_array_2d, self.replace_complex_array_2d)
			# types.STRING2D:    (self.get_string_array_2d,  self.put_string_array_2d,  self.replace_string_array_2d)
				 }.get(code)
		if T is not None:
			return T[method_type]
		return None


	def _method_for_value(self, value, method_type):
		if isinstance(value, np.float32) or isinstance(value, np.float64):
			value = float(value)
		if isinstance(value, np.int32) or isinstance(value, np.int64):
			value = int(value)

		if isinstance(value, str):
			method = (self.get_string,  self.put_string,  self.replace_string)
			return method[method_type]

		T = type(value)

		method = self._method_for_type(T, method_type)
		if method: 
			return method
		if hasattr(value,'__len__'):
			#let numpy work out what type this should be.
			array = np.array(value)
			kind = array.dtype.kind
			ndim = array.ndim
			if ndim==1:
				#1D arrays have their own specific methods,
				#for integer and float
				if kind=='i':
					method = (self.get_int_array_1d,self.put_int_array_1d,self.replace_int_array_1d)
				elif kind=='f':
					method = (self.get_double_array_1d,self.put_double_array_1d,self.replace_double_array_1d)
				elif kind == 'U':
					method = (self.get_string_array_1d,self.put_string_array_1d,self.replace_string_array_1d)
			#otherwise we just use the generic n-d arrays
			elif kind=='i':
				method = (self.get_int_array_nd,self.put_int_array_nd,self.replace_int_array_nd)
			elif kind=='f':
				method = (self.get_double_array_nd,self.put_double_array_nd,self.replace_double_array_nd)
			if method:
				return method[method_type]
		raise ValueError("I do not know how to handle this type %r %r"%(value,type(value)))
	
	def get(self, section, name):
		u"""Get the value of parameter with `name` in `section`.

		The type value returned from this method will reflect the type of
		value held in the underlying map implementation.  In circumstances
		where this either cannot be ascertained or cannot be converted
		simply to a native Python type, then either a :class:`BlockError`
		or :class:`ValueError` will be raised.

		"""
		type_code_c = lib.c_datatype()
		status = lib.c_datablock_get_type(self._ptr, section.encode('ascii'), name.encode('ascii'), ct.byref(type_code_c))
		if status:
			raise BlockError.exception_for_status(status, section, name)
		type_code = type_code_c.value
		method = self._method_for_datatype_code(type_code,self.GET)
		if method:
			return method(section,name)
		raise ValueError("Cosmosis internal error; unknown type of data. section: %s, name: %s, type_code: %s" % (section, name, type_code))

	def put(self, section, name, value, **meta):
		u"""Add a parameter with `value` at (`section`, `name`) in the map.

		The parameter stored in the map will have a type which
		reflects the type of `value`.

		If provided, `meta` should be a map of key/value pairs, and
		these will be appended to the inserted parameter as meta-data,
		converted to string type.

		It is an error to insert a parameter when there already is an
		entry at (`section`, `name`), in which case a :class:`BlockError`
		specialization will be raised.

		"""
		method = self._method_for_value(value,self.PUT)
		method(section, name, value)
		for (key, val) in list(meta.items()):
			self.put_metadata(section, name, str(key), str(val))

	def replace(self, section, name, value):
		u"""Replace the value of a parameter at (`section`, `name`) in the map with `value`.

		The parameter newly stored in the map will have a type which
		reflects the type of `value`.

		It is an error to attempt to replace a parameter not already
		present in the map, in which case a :class:`BlockError`
		specialization will be raised.

		"""
		method = self._method_for_value(value,self.REPLACE)
		method(section, name, value)


	def replace_int(self, section, name, value):
		u"""Change the value of an integer parameter in the map.

		The parameter at (`section`, `name`) will be given the new
		`value`.  It is an error to attempt to replace a value which is
		not already in the map, and a :class:`BlockError` will be raised
		in this case.

		"""
		status = lib.c_datablock_replace_int(self._ptr,section.encode('ascii'),name.encode('ascii'),value)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def replace_bool(self, section, name, value):
		u"""Change the value of a boolean parameter in the map.

		The parameter at (`section`, `name`) will be given the new
		`value`.  It is an error to attempt to replace a value which is
		not already in the map, and a :class:`BlockError` will be raised
		in this case.

		"""
		status = lib.c_datablock_replace_bool(self._ptr,section.encode('ascii'),name.encode('ascii'),value)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def replace_double(self, section, name, value):
		u"""Change the value of a floating-point parameter in the map.

		The parameter at (`section`, `name`) will be given the new
		`value`.  It is an error to attempt to replace a value which is
		not already in the map, and a :class:`BlockError` will be raised
		in this case.

		"""
		r = ct.c_double()
		status = lib.c_datablock_replace_double(self._ptr,section.encode('ascii'),name.encode('ascii'),value)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def replace_complex(self, section, name, value):
		u"""Change the value of a complex parameter in the map.

		The parameter at (`section`, `name`) will be given the new
		`value`.  It is an error to attempt to replace a value which is
		not already in the map, and a :class:`BlockError` will be raised
		in this case.

		"""
		value=self.python_to_c_complex(value)
		status = lib.c_datablock_replace_complex(self._ptr,section.encode('ascii'),name.encode('ascii'),value)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def replace_string(self, section, name, value):
		u"""Change the value of a string parameter in the map.

		The parameter at (`section`, `name`) will be given the new
		`value`.  It is an error to attempt to replace a value which is
		not already in the map, and a :class:`BlockError` will be raised
		in this case.

		"""
		status = lib.c_datablock_replace_string(self._ptr,section.encode('ascii'),name.encode('ascii'),str(value).encode('ascii'))
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def replace_int_array_1d(self, section, name, value):
		u"""Replace the value of a parameter with a simple integer array.

		The parameter at (`section`, `name`) is replaced with `value`,
		interpreted as a one-dimensional array.

		If this cannot be done then a :class:`BlockError` specialization
		will be raised.

		"""
		value_ref, value,n=self.python_to_1d_c_array(value, np.intc)
		status = lib.c_datablock_replace_int_array_1d(self._ptr, section.encode('ascii'), name.encode('ascii'), value, n)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def replace_double_array_1d(self, section, name, value):
		u"""Replace the value of a parameter with a simple floating-point array.

		The parameter at (`section`, `name`) is replaced with `value`,
		interpreted as a one-dimensional array.

		If this cannot be done then a :class:`BlockError` specialization
		will be raised.

		"""
		value_ref, value,n=self.python_to_1d_c_array(value, np.double)
		status = lib.c_datablock_replace_double_array_1d(self._ptr, section.encode('ascii'), name.encode('ascii'), value, n)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def replace_string_array_1d(self, section, name, value):
		u"""Replacing string arrays is not yet implemented

		"""
		raise NotImplementedError("CosmoSIS cannot yet replace string arrays. "
								  "Please open an issue if you need this feature")


	def has_section(self, section):
		u"""Indicate whether or not there is a given `section` in the data set.

		The `section` should be a string holding the name of the section.

		"""
		has = lib.c_datablock_has_section(self._ptr, section.encode('ascii'))
		return bool(has)

	def has_value(self, section, name):
		u"""Indicate whether or not a parameter is in the map.

		Both `section` and `name` should be strings.

		"""
		has = lib.c_datablock_has_value(self._ptr, section.encode('ascii'), name.encode('ascii'))
		return bool(has)

	def __getitem__(self, section_name):
		u"""Get the value of a parameter with `section`, name in the tuple section_name.

		Implicit use of this method is the recommended way to get the
		value of a parameter in the map.  If the argument cannot be
		interpreted as at least a two-item tuple then a
		:class:`ValueError` will be raised.

		"""
		try:
			(section,name) = section_name
		except ValueError:
			raise ValueError("You must specify both a section and a name to get or set a block item: b['section','name']")
		return self.get(section, name)

	def __setitem__(self, section_name, value):
		u"""Set a parameter with value in the map.

		The section_name must be a tuple with the parameterʼs section and
		name as the first two items (else a :class:`ValueError` will be
		raised).  Implicit use of this method is the recommended way to
		both insert parameters into the map, or to introduce new ones.

		"""

		try:
			(section,name) = section_name
		except ValueError:
			raise ValueError("You must specify both a section and a name to get or set a block item: b['section','name']")
		if self.has_value(section, name):
			self.replace(section, name, value)
		else:
			self.put(section, name, value)

	def __contains__(self, section_name):
		u"""Indicate whether there is a parameter with given section/name in the database.

		The section and name must be specified as the first two items of a
		tuple, or else a :class:`ValueError` will be raised.  Normally a
		boolean is returned to indicate prescence of said parameter.
		Implicit use of this method is the recommended way to determine if
		a parameter is present in the data set.

		"""
		if isinstance(section_name, str):
			return self.has_section(section_name)
		try:
			(section,name) = section_name
		except ValueError:
			raise ValueError("You must specify both a section and a name to get or set a block item: b['section','name']")
		return self.has_value(section, name)

	def sections(self):
		u"""Return a list of strings with the names of all sections in the data set.

		"""
		n = lib.c_datablock_num_sections(self._ptr)
		return [lib.c_datablock_get_section_name(self._ptr, i).decode('utf-8') for i in range(n)]


	def keys(self, section=None):
		u"""Return all keys in the collection, or, if `section` is specified, all keys under that section.

		If `section` is specified, it must be a string naming a section
		for whose keys are requested.

		In all cases a list of pairs of strings will be returned, the
		elements of each being the `section` and name of each parameter.

		"""
		if section is None:
			sections = self.sections()
		else:
			sections = [section]
		keys = []
		for section in sections:
			n_value = lib.c_datablock_num_values(self._ptr, section.encode('ascii'))
			for i in range(n_value):
				name = lib.c_datablock_get_value_name(self._ptr, section.encode('ascii'), i).decode('utf-8')
				keys.append((section,name))
		return keys


	def _delete_section(self, section):
		"Internal use only!"
		status = lib.c_datablock_delete_section(self._ptr, section.encode('ascii'))
		if status!=0:
			raise BlockError.exception_for_status(status, section, "<tried to delete>")

	def _copy_section(self, source, dest):
		"Internal use only!"
		status = lib.c_datablock_copy_section(self._ptr, source.encode('ascii'), dest.encode('ascii'))
		if status!=0:
			raise BlockError.exception_for_status(status, dest, "<tried to copy>")


	@staticmethod
	def _parse_metadata_key(key):
		key = key[len(metadata_prefix):].strip(":")
		s = key.index(":")
		if s==-1:
			raise ValueError("Could not understand metadata")
		name = key[:s]
		meta = key[s+1:]
		return name, meta

	def save_to_file(self, dirname, clobber=False):
		u"""Effectively :func:`save_to_directory` with the result tarʼd and compressed to a single file.

		The `dirname` argument here is actually a file name without an
		extension; the path to the file will be created in the file system
		if necessary (:class:`ValueError` will be raised if this cannot be
		accomplished), and “.tgz” will be appended to the file name.

		"""
		filename = dirname + ".tgz"

		base_dirname,base_filename=os.path.split(filename)
		if base_dirname:
			try:
				mkdir(base_dirname)
			except OSError:
				pass

		if os.path.exists(filename) and not clobber:
			raise ValueError("File %s already exists and not clobbering"%filename)

		tar = tarfile.open(filename, "w:gz")

		for (section, scalar_outputs, vector_outputs, meta) in self._save_paths():
			#Save all the vector outputs as individual files
			for name, value in vector_outputs:
				vector_outfile = os.path.join(dirname,section,name+'.txt')
				header = "%s\n"%name
				if value.ndim>2:
					header += "shape = %s\n"%str(value.shape)
					print("Flattening %s--%s when saving; shape info in header" % (section,name))
					value = value.flatten()
				if name in meta:
					for key,val in list(meta[name].items()):
						header+='%s = %s\n' % (key,val)

				#Save this file into the tar file
				string_output = io.BytesIO()
				if value.dtype.kind == "U":
					np.savetxt(string_output, value, header=header.rstrip("\n"), fmt="%s")
				else:
					np.savetxt(string_output, value, header=header.rstrip("\n"))
				string_output.seek(0)
				info = tarfile.TarInfo(name=vector_outfile)
				info.size=len(string_output.getvalue())
				tar.addfile(tarinfo=info, fileobj=string_output)

			#Save all the scalar outputs together as a single file
			#inside the tar file
			if scalar_outputs:
				scalar_outfile = os.path.join(dirname,section,"values.txt")
				string_output = io.BytesIO()
				for s in scalar_outputs:
					line = "{} = {}\n".format(s[0], s[1])
					string_output.write(line.encode())
					if s[0] in meta:
						for key,val in list(meta[s[0]].items()):
							line = "#{} {} = {}\n".format(s[0], key, val)
							string_output.write(line.encode())
				string_output.seek(0)
				info = tarfile.TarInfo(name=scalar_outfile)
				info.size=len(string_output.getvalue())
				tar.addfile(tarinfo=info, fileobj=string_output)
		tar.close()


	def save_to_directory(self, dirname, clobber=False):
		u"""Save the entire contents of this parameter map in the filesystem under `dirname`.

		The data are all written out long-hand in ASCII.  Each unique
		section will go to its own sub-directory, in which all the
		scalar parameters in that section go into a single file
		(‘values.txt’), and of the ‘composite’ data each go into their
		own file, named after the parameter key.

		The path, including `dirname`, will be created if necessary.

		"""
		try:
			mkdir(dirname)
		except OSError:
			if not clobber:
				print("Not clobbering", clobber)
				raise
		for (section, scalar_outputs, vector_outputs, meta) in self._save_paths():
			
			#Create the sub-directory for this 
			#section
			try:
				mkdir(os.path.join(dirname,section))
			except OSError:
				if not clobber:
					raise

			#Save all the vector outputs as individual files
			for name, value in vector_outputs:
				vector_outfile = os.path.join(dirname,section,name+'.txt')
				header = "%s\n"%name
				if value.ndim>2:
					header += "shape = %s\n"%str(value.shape)
					print("Flattening %s--%s when saving; shape info in header" % (section,name))
					value = value.flatten()
				if name in meta:
					for key,val in list(meta[name].items()):
						header+='%s = %s\n' % (key,val)
				if value.dtype.kind == "U":
					np.savetxt(vector_outfile, value, header=header.rstrip("\n"), fmt='%s')
				else:
					np.savetxt(vector_outfile, value, header=header.rstrip("\n"))

			#Save all the scalar outputs together as a single file
			if scalar_outputs:
				f=open(os.path.join(dirname,section,"values.txt"), 'w')
				for s in scalar_outputs:
					f.write("%s = %r\n"%s)
					if s[0] in meta:
						for key,val in list(meta[s[0]].items()):
							f.write("#%s %s = %s\n"%(s[0],key,val))
				f.close()

	def _save_paths(self):
		keys = list(self.keys())
		sections = set(k[0] for k in keys)
		for section in sections:
			scalar_outputs = []
			meta = collections.defaultdict(dict)
			vector_outputs = []
			for k in keys:
				sec, name = k
				if sec!=section: continue
				if name.startswith(metadata_prefix):
					target, metakey = self._parse_metadata_key(name)
					meta[target][metakey] = self[section,name]
					continue
				value = self[section,name]
				if np.isscalar(value):
					scalar_outputs.append((name,value))
				else:
					vector_outputs.append((name,value))
			yield section, scalar_outputs, vector_outputs, meta


	def report_failures(self):
		u"""Dump a human-readable list of failed-action log entries to the standard error channel.

		The entries appear one per line, with space-separated items
		corresponding to the verb, section and name, and data-type of the
		parameter.

		"""
		status = lib.c_datablock_report_failures(self._ptr)
		if status!=0:
			raise BlockError.exception_for_status(status, "", "")

	def print_log(self):
		u"""Dump a human-readable list of log entries to standard output.

		The entries appear one per line, with space-separated items
		corresponding to the verb, section and name, and data-type of the
		parameter.

		"""
		status = lib.c_datablock_print_log(self._ptr)
		if status!=0:
			raise BlockError.exception_for_status(status, "", "")

	def get_log_count(self):
		u"""Return the number of entries in the log."""
		return lib.c_datablock_get_log_count(self._ptr)

	def get_log_entry(self, i):
		u"""Get the iʼth log entry.

		The return is a tuple of four strings indicating the verb (i.e.,
		logged action), section and name of the parameter, and the data
		type held by the parameter.

		"""
		smax = 128
		ptype = ct.create_string_buffer(smax)
		section = ct.create_string_buffer(smax)
		name = ct.create_string_buffer(smax)
		dtype = ct.create_string_buffer(smax)
		status = lib.c_datablock_get_log_entry(self._ptr, i, smax, ptype, section, name, dtype)
		if status:
			raise ValueError("Asked for log entry above maximum or less than zero")
		return ptype.value.decode('utf-8'), section.value.decode('utf-8'), name.value.decode('utf-8'), dtype.value.decode('utf-8')

	def log_access(self, log_type, section, name):
		u"""Add an entry to the end of this :class:`DataBlock` access log.

		The `log_type` describes the action performed on the parameter at
		(`section`, `name`).  It should be one of the strings displayed in
		*datablock_logging.cc*, viz: "READ-OK", "WRITE-OK", "READ-FAIL",
		"WRITE-FAIL", "READ-DEFAULT", "REPLACE-OK", "REPLACE-FAIL",
		"CLEAR", "DELETE", or "MODULE-START".

		"""
		status = lib.c_datablock_log_access(self._ptr, log_type.encode('ascii'), section.encode('ascii'), name.encode('ascii'))
		if status!=0:
			raise BlockError.exception_for_status(status, "", "")

	def get_metadata(self, section, name, key):
		u"""Get the metadata called `key` attached to parameter `name` under `section`.

		If the data do not exist at the requested address, then a
		:class:`BlockError` will be raised.

		"""
		r = lib.c_str()
		status = lib.c_datablock_get_metadata(self._ptr,section.encode('ascii'),name.encode('ascii'),key.encode('ascii'), r)
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)
		return r.value.decode('utf-8')

	def put_metadata(self, section, name, key, value):
		u"""Associate `value` with the meta-`key` attached to parameter `name` under `section`.

		If there is no parameter under (`section`, `name`) then a
		:class:`BlockError` will be raised.

		"""
		status = lib.c_datablock_put_metadata(self._ptr,section.encode('ascii'),name.encode('ascii'),key.encode('ascii'), value.encode('ascii'))
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def replace_metadata(self, section, name, key, value):
		u"""Associate `value` with the meta-`key` attached to parameter `name` under `section`.

		If there is no parameter under (`section`, `name`) then a
		:class:`BlockError` will be raised.

		"""
		status = lib.c_datablock_replace_metadata(self._ptr,section.encode('ascii'),name.encode('ascii'),key.encode('ascii'), value.encode('ascii'))
		if status!=0:
			raise BlockError.exception_for_status(status, section, name)

	def put_grid(self, section, name_x, x, name_y, y, name_z, z):
		u"""Put a grid into the map.

		The grid is put into `section`, using keys `name_x`, `name_y` and
		`name_z` to locate the data.  The data comprise the array `x`
		holding a set of ‘labels’ for the x-axis, an array `y` holding
		labels for the y-axis, and then a two-dimensional array`z`, whose
		sizes must correspond with the `x`- and `y`-sizes, which holds the
		actual data inside the grid.

		If there are any problems, most notably with the sizes of the
		arrays not being compatible, then a :class:`ValueError` will be
		raised.

		"""
		self._grid_put_replace(section, name_x, x, name_y, y, name_z, z, False)

	def get_grid(self, section, name_x, name_y, name_z):
		u"""Return a triple of arrays, representing a grid of data.

		The strings `name_x`, `name_y` and `name_z` must be keys under
		`section` which index data making up a grid; they must be the same
		set used in a call to :func:`replace_grid` or :func:`put_grid`
		used to establish the grid in the first place (except that the x-
		and y-axes are allowed to be transposed).

		The return is a triple of arrays: the first two elements hold the
		labels along the axes and the third element is a two-dimensional
		array holding the data deemed to be inside the grid itself.

		If the `name_*`ʼs do not correspond correctly with those of an
		established grid then a :class:`BlockError` will be raised.

		"""
		name_x = name_x.lower()
		name_y = name_y.lower()
		name_z = name_z.lower()
		x = self[section, name_x]
		y = self[section, name_y]
		z = self[section, name_z]
		sentinel_key = "_cosmosis_order_%s"%name_z
		sentinel_value = self[section, sentinel_key].lower()

		if sentinel_value== "%s_cosmosis_order_%s" % (name_x, name_y):
			assert z.shape==(x.size, y.size)
			return x, y, z
		elif sentinel_value== "%s_cosmosis_order_%s" % (name_y, name_x):
			assert z.T.shape==(x.size, y.size)
			return x, y, z.T
		else:
			raise BlockError.exception_for_status(errors.DBS_WRONG_VALUE_TYPE, section, name_z)



	def replace_grid(self, section, name_x, x, name_y, y, name_z, z):
		u"""Put a grid into the map.

		The grid is put into `section`, using keys `name_x`, `name_y` and
		`name_z` to locate the data.  The data comprise the array `x`
		holding a set of ‘labels’ for the x-axis, an array `y` holding
		labels for the y-axis, and then a two-dimensional array`z`, whose
		sizes must correspond with the `x`- and `y`-sizes, which holds the
		actual data inside the grid.

		If there are any problems, most notably with the sizes of the
		arrays not being compatible, then a :class:`ValueError` will be
		raised.

		"""
		self._grid_put_replace(section, name_x, x, name_y, y, name_z, z, True)

	def _grid_put_replace(self, section, name_x, x, name_y, y, name_z, z, replace):
		# These conversions do not create new objects if x,y,z are already arrays.
		x = np.asarray(x)
		y = np.asarray(y)
		z = np.asarray(z)

		if x.ndim!=1 or y.ndim!=1 or z.ndim!=2 or z.shape!=(x.size,y.size):
			msg = """
	Your code tried to save or replace a grid {name_z}[{name_x}, {name_y}] in section {}. 
	This requires 1D {name_x}, 1D {name_y}, 2D {name_z} and the shape of {name_z} to be (len({name_x}),len({name_y})).
	Whereas your code tried:
	{name_x} ndim = {}    [{}]
	{name_y} ndim = {}    [{}]
	{name_z} ndim = {}    [{}]
	{name_x} shape = {}
	{name_y} shape = {}
	{name_z} shape = {}   [{}]
			""".format(section, 
				x.ndim, "OK" if x.ndim==1 else "WRONG",
				y.ndim, "OK" if y.ndim==1 else "WRONG",
				z.ndim, "OK" if z.ndim==2 else "WRONG",
				x.shape,
				y.shape,
				z.shape, "OK" if z.shape==(x.size, y.size) else "WRONG",
				name_z=name_z, name_x=name_x, name_y=name_y
				)
			raise ValueError(msg)

		self[section, name_x] = x
		self[section, name_y] = y
		self[section, name_z] = z

		sentinel_key = "_cosmosis_order_%s"%name_z
		sentinel_value = "%s_cosmosis_order_%s" % (name_x, name_y)
		self[section, sentinel_key] = sentinel_value.lower()

	def get_first_parameter_use(self, params_of_interest):
		u"""Analyze the log and figure out when each parameter is first used"""
		params_by_module = collections.OrderedDict()
		current_module = []
		#make a copy of the parameter list so we can remove things
		#from it as we find their first use
		params = [(p.section,p.name) for p in params_of_interest]
		#now actually parse the log
		current_module = None
		current_name = "None"
		for i in range(self.get_log_count()):
			ptype, section, name, _ = self.get_log_entry(i)
			if ptype=="MODULE-START":
				# The previous current_module is already the
				#last element in params_by_module (unless it's the
				#very first one in which case we discard it because
				#it is the parameters being set in the sampler)
				current_module = []
				current_name = section
				params_by_module[current_name] = current_module
			elif ptype=="READ-OK" and (section,name) in params:
				current_module.append((section,name))
				params.remove((section,name))
		#Return a list of lists of parameter first used in each section
		return params_by_module
	
	def get_all_parameter_use(self, params_of_interest):
		u"""Analyze the log and figure out which parameter is in use in specific module"""
		params_by_module = collections.OrderedDict()
		current_module = []
		#make a copy of the parameter list
		params = [(p.section,p.name) for p in params_of_interest]
		#now actually parse the log
		current_module = None
		current_name = "None"
		for i in range(self.get_log_count()):
			ptype, section, name, _ = self.get_log_entry(i)
			if ptype=="MODULE-START":
				# The previous current_module is already the
				#last element in params_by_module (unless it's the
				#very first one in which case we discard it because
				#it is the parameters being set in the sampler)
				current_module = []
				current_name = section
				params_by_module[current_name] = current_module
			elif ptype=="READ-OK" and (section, name) in params and (section, name) not in current_module:
				current_module.append((section, name))
		#Return a list of lists of parameter used in each module
		return params_by_module

	@classmethod
	def from_yaml(cls, filename_or_stream):
		import yaml

		block = cls()
		if isinstance(filename_or_stream, str):
			stream = open(filename_or_stream)
		else:
			stream = filename_or_stream

		data = yaml.safe_load(stream)

		if not isinstance(data, dict) or (not all(isinstance(k, str) for k in data.keys())) or (not all(isinstance(v, dict) for v in data.values())):
			raise ValueError("Wrong format yaml file {}".format(filename_or_stream))

		for section, contents in data.items():
			for key, value in contents.items():
				if isinstance(value, list):
					value = np.array(value)
				block[section, key] = value

		return block

	def to_yaml(self, filename_or_stream):
		import yaml
		if isinstance(filename_or_stream, str):
			stream = open(filename_or_stream, 'w')
		else:
			stream = filename_or_stream

		data = {}
		for section in self.sections():
			data[section] = {}
			for (_, key) in self.keys(section):
				value = self[section, key]
				if isinstance(value, np.ndarray):
					value = value.tolist()
				data[section][key] = value

		yaml.dump(data, stream)

	@classmethod
	def from_pickle(cls, filename):
		import pickle

		block = cls()
		with open(filename + ".pkl", 'rb') as f:
			data = pickle.load(f)
		for section, values in data.items():
			for key, value in values.items():
				block[section, key] = value
		return block

	def to_pickle(self, filename):
		import pickle

		data = {}
		for section in self.sections():
			data[section] = {}
			for (_, key) in self.keys(section):
				value = self[section, key]
				data[section][key] = value

		with open(filename + ".pkl", 'wb') as f:
			pickle.dump(data, f)

	@classmethod
	def from_dict(cls, d):
		b = cls()
		for section, values in d.items():
			for key, value in values.items():
				b[section, key] = value
		return b

	@classmethod
	def from_string(cls, s):
		sio = StringIO(s)
		return cls.from_yaml(sio)

	def to_string(self):
		sio = StringIO()
		self.to_yaml(sio)
		sio.seek(0)
		return sio.read()

	def __reduce__(self):
		return (DataBlock.from_string, (self.to_string(),))








class SectionOptions(object):
	"""
	The SectionOptions object wraps is a handy short-cut to let you
	look up objects in a DataBlock object, but looking specifically at
	the special section in an ini file that refers to "the section that 
	defines the current module"

	"""
	def __init__(self, block):
		self.block=block

	def has_value(self, name):
		has = self.block.has_value(option_section, name)
		return bool(has)



def _make_getter(cls, name):
	if name=='__getitem__':
		def getter(self, key):
			return self.block[option_section, key]
	elif "array" in name:
		def getter(self, key):
			return getattr(self.block, name)(option_section, key)
	else:
		def getter(self, key, default=None):
			return getattr(self.block, name)(option_section, key, default=default)

	return getter




for name in dir(DataBlock):
	if name.startswith('get') or name=='__getitem__':
		setattr(SectionOptions, name, _make_getter(SectionOptions, name))




try:
	from cosmosis import names as section_names
except ImportError:
	print("Running without cosmosis: no pretty section names")
	section_names = None


class NoSuchParameter(Exception):
	pass

def section_code(code):
	if section_names:
		for att in dir(section_names):
			if getattr(section_names,att)==code:
				return att
	return code




from .. import ParallelSampler
from ...runtime import logs
import numpy as np
import ctypes as ct
import os
import sys

MINUIT_INI_SECTION = "minuit"

loglike_type = ct.CFUNCTYPE(
    ct.c_double, #likelihood
    ct.POINTER(ct.c_double),  #parameter cube
)

class MinuitOptionsStruct(ct.Structure):
    _fields_ = [
        ("max_evals", ct.c_int),
        ("strategy", ct.c_int),
        ("algorithm", ct.c_int),
        ("save_cov", ct.c_char_p),
        ("tolerance", ct.c_double),
        ("width_estimate", ct.c_double),
        ("do_master_output", ct.c_int),
    ]

libname=os.path.join(os.path.split(__file__)[0],"minuit_wrapper.so")

class MinuitSampler(ParallelSampler):
    needs_output = True
    needs_parallel_output = False
    libminuit_name = libname
    libminuit = None
    sampler_outputs = [("prior", float), ("post", float)]

    def config(self):
        self.converged = False
        if MinuitSampler.libminuit is None:
            if not os.path.exists(self.libminuit_name):
                raise ValueError("The CosmoSIS minuit2 wrapper was not compiled.  If you installed CosmoSIS manually you need the library minuit2 to use this sampler. See the wiki for more details.")
            MinuitSampler.libminuit = ct.cdll.LoadLibrary(self.libminuit_name)
        self.maxiter = self.read_ini("maxiter", int, 1000)
        self.save_dir = self.read_ini("save_dir", str, "")
        self.save_cov = self.read_ini("save_cov", str, "")
        self.output_ini = self.read_ini("output_ini", str, "")
        self.verbose = self.read_ini("verbose", bool, False)
        self.iterations = 0

        #Minuit options
        self.strategy = self.read_ini("strategy", str, "medium").lower()
        self.algorithm = self.read_ini("algorithm", str, "migrad").lower()

        self.width_estimate = self.read_ini("width_estimate", float, 0.05)
        self.tolerance = self.read_ini("tolerance", float, 50.0)
        self.neval = 0
        self.param_vector = None

        strategy = {
            "fast":0,
            "medium":1,
            "safe":2,
            "0":0,
            "1":1,
            "2":2,
        }.get(self.strategy)
        if strategy is None:
            raise ValueError("Minuit 'strategy' parameter must be one of fast, medium, safe (default=medium) not '%s'"%self.strategy)
        self.strategy = strategy

        algorithm = {
            "migrad":0,
            "fallback":1,
            "combined":1,
            "simplex":2,
            # "scan":3,
        }.get(self.algorithm)
        if algorithm is None:
            raise ValueError("Minuit 'algorithm' parameter must be one of migrad, fallback, simplex (default=migrad) not '%s'"%strategy)
        self.algorithm = algorithm


        self._run = MinuitSampler.libminuit.cosmosis_minuit2_wrapper
        self._run.restype = ct.c_int
        vec = ct.POINTER(ct.c_double)
        self.ndim = len(self.pipeline.varied_params)

        cube_type = ct.c_double*self.ndim

        @loglike_type
        def wrapped_likelihood(cube_p):
            vector = np.frombuffer(cube_type.from_address(ct.addressof(cube_p.contents)))
            try:
                like, extra = self.pipeline.posterior(vector)
            except KeyboardInterrupt:
                sys.exit(1)
            self.iterations += 1
            return -like

        self.wrapped_likelihood = wrapped_likelihood


    def execute(self):
        if self.param_vector is None:
            self.param_vector = self.start_estimate()

        #Run an iteration of minuit
        param_vector, param_names, results, status, made_cov, cov_vector = self.sample()


        #update the current parameters
        self.param_vector = param_vector.copy()
        self.neval += status

        if status == 0:
            logs.overview("SUCCESS: Minuit has converged!")
            self.save_results(param_vector, param_names, results)
            self.converged = True
            self.distribution_hints.set_peak(param_vector, results.post)

            if made_cov:
                cov_matrix = cov_vector.reshape((self.ndim, self.ndim))
                self.distribution_hints.set_cov(cov_matrix)
        elif self.neval > self.maxiter:
            logs.error("MINUIT has failed to converge properly in the max number of iterations.  Sorry.")
            logs.error("Saving the best fitting parameters of the ones we tried, though beware: these are probably not the best-fit")
            self.save_results(param_vector, param_names, results)
            #we actually just use self.converged to indicate that the 
            #sampler should stop now
            self.converged = True

        else:
            logs.overview("Minuit did not converge this run; trying again")
            logs.overview("until we run out of iterations.")
            


    def save_results(self, param_vector, param_names, results):
        section = None

        if self.pool is not None:
            logs.error("Note that the # of function calls printed above is not the total count for all cores, just for one core.")

        self.output.parameters(param_vector, results.extra, results.prior, results.post)

        if self.output_ini:
            self.pipeline.create_ini(param_vector, self.output_ini)

        for name, value in zip(param_names, param_vector):
            sec,name=name.split('--')
            if section!=sec:
                logs.important("[%s]" % sec)
                section=sec
            logs.important("%s = %g" % (name,value))

        logs.important(f"Likelihood = {results.like}")

        if self.save_dir:
            logs.overview(f"Saving best-fit model cosmology to {self.save_dir}")
            results.block.save_to_directory(self.save_dir, clobber=True)

        self.converged = True

    def sample(self):

        param_names = [str(p) for p in self.pipeline.varied_params]
        param_names_array = (ct.c_char_p * len(param_names))()
        param_names_array[:] = [p.encode('ascii') for p in param_names]

        master = 1 if self.pool is None or self.pool.is_master() else 0
        
        param_max = self.pipeline.max_vector()
        param_min = self.pipeline.min_vector()

        param_vector = self.param_vector.copy()

        cov_vector = np.zeros(self.ndim*self.ndim)
        made_cov = ct.c_int()

        options = MinuitOptionsStruct(
            #allow for more loops
            max_evals=self.maxiter-self.neval, strategy=self.strategy, 
            algorithm=self.algorithm, save_cov=self.save_cov.encode('ascii'),
            tolerance=self.tolerance, width_estimate=self.width_estimate,
            do_master_output = master
            )

        status = self._run(self.ndim, 
            param_vector.ctypes.data_as(ct.POINTER(ct.c_double)), 
            param_min.ctypes.data_as(ct.POINTER(ct.c_double)), 
            param_max.ctypes.data_as(ct.POINTER(ct.c_double)), 
            self.wrapped_likelihood, 
            param_names_array,
            cov_vector.ctypes.data_as(ct.POINTER(ct.c_double)),
            ct.byref(made_cov),
            options
            )

        #Run the pipeline one last time ourselves, so we can save the 
        #likelihood and cosmology
        results = self.pipeline.run_results(param_vector)
        return param_vector, param_names, results, status, made_cov, cov_vector

    def worker(self):
        self.sample()


    def is_converged(self):
        return self.converged



from .. import ParallelSampler, sample_ball
from ...runtime import logs
import numpy as np

pipeline = None
def log_probability_function(p):
    global pipeline
    r = pipeline.run_results(p)
    return r.post, (r.extra, r.prior)




class KombineSampler(ParallelSampler):
    parallel_output = False
    sampler_outputs = [("prior", float), ("post", float)]

    def config(self):
        global pipeline
        pipeline = self.pipeline

        if self.is_master():
            try:
                import kombine
            except ImportError:
                raise ImportError("The kombine package must be installed to use the kombine sampler\n" +
                            "You can use pip from inside the CosmoSIS environment:\n" +
                            "pip install --user git+git://github.com/bfarr/kombine")
            self.kombine = kombine

            # Parameters of the kombine sampler
            self.nwalkers = self.read_ini("walkers", int, 2)
            self.samples = self.read_ini("samples", int, 1000)
            self.nsteps = self.read_ini("nsteps", int, 100)
            self.update_interval = self.read_ini("update_interval", int, 100)
            start_file = self.read_ini("start_points", str, "")
            random_start = self.read_ini("random_start", bool, False)

            #Starting positions and values for the chain
            self.ndim = len(self.pipeline.varied_params)
            self.num_samples = 0
            self.lnpost0 = None
            self.lnprop0 = None
            self.blob0 = None

            if start_file:
                self.p0 = self.load_start(start_file)
                print("Loaded starting position from %s", start_file)
            elif random_start:
                self.p0 = [self.pipeline.randomized_start()
                           for i in range(self.nwalkers)]
            else:
                center_norm = self.pipeline.normalize_vector(self.pipeline.start_vector())
                sigma_norm=np.repeat(1e-3, center_norm.size)
                p0_norm = sample_ball(center_norm, sigma_norm, size=self.nwalkers)
                self.p0 = [self.pipeline.denormalize_vector(p0_norm_i) for p0_norm_i in p0_norm]

            #Finally we can create the sampler
            self.ensemble = self.kombine.Sampler(self.nwalkers, self.ndim,
                                                 log_probability_function,
                                                 processes=1,
                                                 pool=self.pool)
            self.burned_in = False

    def load_start(self, filename):
        #Load the data and cut to the bits we need.
        #This means you can either just use a test file with
        #starting points, or an output file.
        data = np.genfromtxt(filename, invalid_raise=False)[-self.nwalkers:, :self.ndim]
        if data.shape != (self.nwalkers, self.ndim):
            raise RuntimeError("There are not enough lines or columns "
                               "in the starting point file %s" % filename)
        return list(data)

    def output_samples(self, pos, prob, extra_info):
        for p,l,e in zip(pos,prob,extra_info):
            e, pr = e
            self.output.parameters(p, e, pr, l)

    def execute(self):
        if not self.burned_in:
            results = self.ensemble.burnin(self.p0, update_interval=self.update_interval)
            try:
                pos, post, prop, extra_info = results
            except ValueError:
                pos, post, prop = results
                extra_info = None
            if self.is_master():
                logs.overview("Burn-in phase complete.")
            self.p0 = pos
            self.lnpost0 = post
            self.lnprop0 = prop
            self.blob0 = extra_info
            self.burned_in = True
            return

        outputs = []
        # unlike emcee, kombine only returns extra_info if posterior function provides it
        for results in self.ensemble.sample(
                self.p0, lnpost0=self.lnpost0, lnprop0=self.lnprop0, blob0=self.blob0,
                storechain=False, iterations=self.nsteps, update_interval=self.update_interval):
            try:
                pos, post, prop, extra_info = results
            except ValueError:
                pos, post, prop = results
                extra_info = []
            outputs.append((pos.copy(), post.copy(), prop.copy(), extra_info[:]))

        for (pos, post, prop, extra_info) in outputs:
            self.output_samples(pos, post, extra_info)
            self.distribution_hints.set_from_sample(pos, post)

        # Set the starting positions for the next chunk of samples
        # to the last ones for this chunk
        self.p0 = pos
        self.lnpost0 = post
        self.prob0 = prop
        self.blob0 = extra_info
        self.num_samples += self.nsteps
        print("Done %d iterations", self.num_samples)

    def is_converged(self):
        return self.num_samples >= self.samples



class _close_pool_message(object):
    def __repr__(self):
        return "<Close pool message>"


class _function_wrapper(object):
    def __init__(self, function, callback=None):
        self.function = function
        self.callback = callback


def _error_function(task):
    raise RuntimeError("Pool was sent tasks before being told what "
                       "function to apply.")


class MPIPool(object):
    def __init__(self, debug=False, comm=None):
        try:
            from mpi4py import MPI
            self.MPI = MPI
        except ImportError:
            raise RuntimeError("MPI environment not found!")
        if comm is None:
            comm = self.MPI.COMM_WORLD
        self.comm = comm
        self.rank = self.comm.Get_rank()
        self.size = self.comm.Get_size()
        self.debug = debug

        self.function = _error_function
        self.callback = None

    def is_master(self):
        return self.rank == 0

    def wait(self):
        if self.is_master():
            raise RuntimeError("Master node told to await jobs")
        status = self.MPI.Status()
        while True:
            task = self.comm.recv(source=0, tag=self.MPI.ANY_TAG,
                                  status=status)

            if isinstance(task, _close_pool_message):
                break

            if isinstance(task, _function_wrapper):
                self.function = task.function
                self.callback = task.callback
                continue

            if self.callback:
                def compose(x):
                    result = self.function(x)
                    self.callback(x, result)
                    return result
                results = list(map(compose, task))
            else:
                results = list(map(self.function, task))
            self.comm.send(results, dest=0, tag=status.tag)

    def map(self, function, tasks, callback=None):
        # Should be called by the master only
        if not self.is_master():
            self.wait()
            return

        tasks= list(tasks)
        # send function if necessary
        if function is not self.function or callback is not self.callback:
            self.function = function
            self.callback = callback
            F = _function_wrapper(function, callback)
            requests = [self.comm.send(F, dest=i)
                        for i in range(1, self.size)]
            #self.MPI.Request.waitall(requests)

        # distribute tasks to workers
        requests = []
        for i in range(1, self.size):
            req = self.comm.send(tasks[i::self.size], dest=i)
            requests.append(req)

        # process local work
        results = [None]*len(tasks)

        if self.callback:
            def compose(x):
                result = self.function(x)
                self.callback(x, result)
                return result
            results[::self.size] = list(map(compose, tasks[::self.size]))
        else:
            results[::self.size] = list(map(self.function, tasks[::self.size]))

        # recover results from workers (in any order)
        status = self.MPI.Status()
        for i in range(self.size-1):
            result = self.comm.recv(source=self.MPI.ANY_SOURCE,
                                    status=status)
            results[status.source::self.size] = result
        return results

    def gather(self, data, root=0):
        return self.comm.gather(data, root)

    def bcast(self, data, root=0):
        return self.comm.bcast(data, root)

    def send(self, data, dest=0, tag=0):
        self.comm.send(data, dest, tag)

    def recv(self, source=0, tag=0):
        return self.comm.recv(source, tag)

    def allreduce(self, data):
        return self.comm.allreduce(data)

    def close(self):
        if self.is_master():
            for i in range(1, self.size):
                self.comm.isend(_close_pool_message(), dest=i)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

class MPILogFile:
    def __init__(self, filename, pool=None):
        if hasattr(pool, "comm"):
            self.comm = pool.comm
        elif pool is None:
            self.comm = None
        else:
            raise ValueError("Failure logging from multiple processes does not yet work using multiprocessing with --smp, only MPI using --mpi")

        self.filename = filename
        if self.comm is None:
            self.file = open(filename, "w")
        else:
            from mpi4py import MPI
            self.MPI = MPI
            self.file = MPI.File.Open(self.comm, filename, MPI.MODE_WRONLY | MPI.MODE_CREATE)

    def write(self, message):
        if self.comm is None:
            self.file.write(message)
        else:
            message = message.encode("utf-8")
            self.file.Write_shared(message)

    def __del__(self):
        if self.file is not None:
            self.close()

    def close(self):
        if self.comm is None:
            self.file.close()
        else:
            self.file.Close()
        self.file = None




import ctypes

def activate_segfault_handling():
    try:
        import faulthandler
    except ImportError:
        raise ImportError("You must install the python faulthandler package to use experimental fault handling")
    from ..datablock.cosmosis_py import enable_cosmosis_segfault_handler
    enable_cosmosis_segfault_handler()
    print("Experimental fault handling enabled")


from .output_base import OutputBase
from . import utils
import numpy as np
import os
from glob import glob
from collections import OrderedDict
try:
    import fitsio
except ImportError:
    fitsio = None

comment_indicator = "_cosmosis_comment_indicator_"
final_metadata_indicator = "FINALMETA"
unreserve_indicator = "UNRES"
reserved_keys = [
    "XTENSION",
    "BITPIX",
    "NAXIS",
    "NAXIS1",
    "NAXIS2",
    "PCOUNT",
    "GCOUNT",
    "TFIELDS",
    "TTYPE1",
    "COMMENT",
]

def check_fitsio():
    if fitsio is None:
        raise RuntimeError("You need to have the fitsio library installed to output FITS files. Try running: pip install  --install-option=\"--use-system-fitsio\" git+git://github.com/joezuntz/fitsio")

class FitsOutput(OutputBase):
    FILE_EXTENSION = ".fits"
    _aliases = ["fits"]

    def __init__(self, filename, rank=0, nchain=1, clobber=True):
        super(FitsOutput, self).__init__()

        #If filename already ends in .txt then remove it for a moment
        if filename.endswith(self.FILE_EXTENSION):
            filename = filename[:-len(self.FILE_EXTENSION)]

        if nchain > 1:
            filename = filename + "_{}".format(rank+1)

        self._filename = filename + self.FILE_EXTENSION

        self.filename_base = filename

        check_fitsio()

        self._fits = fitsio.FITS(self._filename, "rw", clobber=clobber)
        self._hdu = None

        #also used to store comments:
        self._metadata = OrderedDict()
        self._final_metadata = OrderedDict()

    def _close(self):
        self._flush_metadata(self._final_metadata)
        self._final_metadata={}
        self._fits.close()

    def _flush_metadata(self, metadata):

        for (key,(value,comment)) in list(metadata.items()):
            if key.startswith(comment_indicator):
                self._hdu.write_comment(value)
            elif comment:
                self._hdu.write_key(key, value, comment)
            else:
                self._hdu.write_key(key, value)


    def _begun_sampling(self, params):
        #write the name line
        self._fits.create_table_hdu(data=params, names=[c[0] for c in self.columns])
        self._hdu = self._fits[-1]
        self._dtype = self._hdu.get_rec_dtype()[0]
        self._flush_metadata(self._metadata)
        self._metadata={}

    @staticmethod
    def is_reserved_fits_keyword(key):
        for k in reserved_keys:
            if key.upper().startswith(k):
                return True
        return False

    def _write_metadata(self, key, value, comment=''):
        #We save the metadata until we get the first 
        #parameters since up till then the columns can
        #be changed
        if self.is_reserved_fits_keyword(key):
            key=unreserve_indicator + key
        self._metadata[key]= (value, comment)

    def _write_comment(self, comment):
        #save comments along with the metadata - nice as 
        #preserves order
        self._metadata[comment_indicator +
                       "_%d" % (len(self._metadata))] = (comment,None)

    def _write_parameters(self, params):
        row = np.core.records.fromarrays(params, dtype=self._dtype)
        row=np.atleast_1d(row)
        self._hdu.append(row)

    def _write_final(self, key, value, comment=''):
        #I suppose we can put this at the end - why not?
        if self.is_reserved_fits_keyword(key):
            key=unreserve_indicator + key
        self._final_metadata[key]= (value, final_metadata_indicator+comment)

    def name_for_sampler_resume_info(self):
        return self.filename_base + '.sampler_status'

    @classmethod
    def from_options(cls, options, resume=False):
        #look something up required parameters in the ini file.
        #how this looks will depend on the ini 
        if resume:
            raise ValueError("Cannot resume from FITS output")
            
        filename = options['filename']
        delimiter = options.get('delimiter', '\t')
        rank = options.get('rank', 0)
        nchain = options.get('parallel', 1)
        clobber = utils.boolean_string(options.get('clobber', True))
        return cls(filename, rank, nchain, clobber=clobber)

    @classmethod
    def load_from_options(cls, options):
        check_fitsio()
        filename = options['filename']

        cut = False
        if filename.endswith(cls.FILE_EXTENSION):
            filename = filename[:-len(cls.FILE_EXTENSION)]
            cut = True
        # first look for serial file
        if os.path.exists(filename+cls.FILE_EXTENSION):
            datafiles = [filename+cls.FILE_EXTENSION]
        elif os.path.exists(filename) and not cut:
            datafiles = [filename]
        else:
            datafiles = glob(filename+"_[0-9]*"+cls.FILE_EXTENSION)
            if not datafiles:
                raise RuntimeError("No datafiles found starting with %s!"%filename)

        #Read the metadata
        metadata = []
        final_metadata = []
        data = []
        comments = []
        column_names = None

        for datafile in datafiles:

            print('LOADING CHAIN FROM FILE: ', datafile)
            chain = []
            chain_metadata = {}
            chain_final_metadata = {}
            chain_comments = []

            f = fitsio.FITS(datafile, "r")
            hdu = f[1]
            chain = f[1].read()
            #convert to unstructured format
            chain = chain.view((chain.dtype[0], len(chain.dtype.names)))

            column_names = hdu.get_colnames()

            hdr = hdu.read_header()
            chain_comments = [r['comment'] for r in hdr.records() if r['name'].lower()=="comment"]
            for r in hdr.records():
                key = r['name']
                if key=='COMMENT':
                    continue
                if key.startswith(unreserve_indicator):
                    key = key[len(unreserve_indicator):]
                value = r['value']
                key=key.lower()
                if r['comment'].startswith(final_metadata_indicator):
                    chain_final_metadata[key] = value
                else:
                    chain_metadata[key] = value

            data.append(np.array(chain))
            metadata.append(chain_metadata)
            final_metadata.append(chain_final_metadata)
            comments.append(chain_comments)

        if column_names is None:
            raise ValueError("Could not find column names header in file starting %s"%filename)

        return column_names, data, metadata, comments, final_metadata
